"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
# This code was pulled from another AIML processor in Github
# https://github.com/Decalogue/aiml3/blob/master/aiml/LangSupport.py
# Thanks to Benny Shang for the info and help supporting chinese
from programy.utils.logging.ylogger import YLogger


class ChineseLanguage:

    @staticmethod
    def is_language(c):
        # http://www.iteye.com/topic/558050

        if isinstance(c, str):
            if len(c) == 0 or len(c) > 1:
                return False

            r = [
                # 标准CJK文字
                (0x3400, 0x4DB5), (0x4E00, 0x9FA5), (0x9FA6, 0x9FBB), (0xF900, 0xFA2D),
                (0xFA30, 0xFA6A), (0xFA70, 0xFAD9), (0x20000, 0x2A6D6), (0x2F800, 0x2FA1D),
                # 全角ASCII、全角中英文标点、半宽片假名、半宽平假名、半宽韩文字母
                (0xFF00, 0xFFEF),
                # CJK部首补充
                (0x2E80, 0x2EFF),
                # CJK标点符号
                (0x3000, 0x303F),
                # CJK笔划
                (0x31C0, 0x31EF)]

            return any(s <= ord(c) <= e for s, e in r)

        return False

    @staticmethod
    def split_language(s):
        result = []
        for c in s:
            if ChineseLanguage.is_language(c):
                result.extend([" ", c, " "])
            else:
                result.append(c)
        ret = ''.join(result)
        return ret.split()

    @staticmethod
    def split_unicode(s):
        segs = s.split()
        result = []
        for seg in segs:
            if any(map(ChineseLanguage.is_language, seg)):
                result.extend(ChineseLanguage.split_language(seg))
            else:
                result.append(seg)
        return result

    @staticmethod
    def get_str_segment(seg, result, num, append):
        str(seg[num]) and result.append(append)

    @staticmethod
    def split_with_spaces(s):
        segs = ChineseLanguage.split_language(s)
        result = []
        for seg in segs:
            # English marks
            if seg[0] not in ".,?!":
                ChineseLanguage.get_str_segment(seg, result, 0, " ")

            result.append(seg)

            ChineseLanguage.get_str_segment(seg, result, -1, " ")

        return ''.join(result).strip()



import json
from urllib.request import urlopen, Request
from packaging.version import Version

def get_current_versions(req_file):

    with open(req_file, "r") as file:
        for line in file:
            line = line.strip()
            if line:
                if line.startswith('#') is False:
                    name_ver = line.split('==')
                    try:
                        url = "https://pypi.python.org/pypi/%s/json" % (name_ver[0])
                        data = json.load(urlopen(Request(url)))
                        versions = list(data["releases"].keys())
                        versions.sort(key=Version)
                        latest = versions[len(versions) - 1]
                        print("Package: %s, current: %s, latest: %s" % (name_ver[0], name_ver[1], latest))
                    except Exception as e:
                        print(name_ver)
                        print(e)

if __name__ == '__main__':

    get_current_versions("../../../requirements.txt")



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from abc import ABC
from abc import abstractmethod
from programy.brain import Brain
from programy.utils.classes.loader import ClassLoader
from programy.utils.logging.ylogger import YLogger


class BrainSelector(ABC):

    def __init__(self, configuration):
        self._configuration = configuration

    @abstractmethod
    def select_brain(self):
        raise NotImplementedError()  # pragma: no cover


class DefaultBrainSelector(BrainSelector):

    def __init__(self, configuration, brains):
        BrainSelector.__init__(self, configuration)
        self._brains = brains
        self._iterator = None
        self._set_iterator()

    def _set_iterator(self):
        self._iterator = iter(self._brains.values())

    def select_brain(self):
        try:
            return next(self._iterator)

        except StopIteration:
            self._set_iterator()

            try:
                return next(self._iterator)
            except StopIteration:
                pass

        return None


class BrainFactory:

    def __init__(self, bot):
        self._brains = {}
        self._brain_selector = None
        self.loads_brains(bot)
        self.load_brain_selector(bot.configuration)

    def brainids(self):
        return list(self._brains.keys())

    def brain(self, brainid):
        if brainid in self._brains:
            return self._brains[brainid]
        else:
            return None

    def loads_brains(self, bot):
        for config in bot.configuration.configurations:
            brain = Brain(bot, config)
            self._brains[brain.id] = brain

    def load_brain_selector(self, configuration):
        if configuration.brain_selector is None:
            self._brain_selector = DefaultBrainSelector(configuration, self._brains)
        else:
            try:
                self._brain_selector = ClassLoader.instantiate_class(configuration.brain_selector)(configuration,
                                                                                                   self._brains)
            except Exception as excep:
                YLogger.exception_nostack(self, "Failed to load defined brain selector, loadiing default", excep)
                self._brain_selector = DefaultBrainSelector(configuration, self._brains)

    def select_brain(self):
        return self._brain_selector.select_brain()

    def get_question_counts(self):
        brains = []
        for brainid, brain in self._brains.items():
            brains.append({"id": brainid,
                           "questions": brain.num_questions})
        return brains



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
from programy.utils.logging.ylogger import YLogger
from programy.extensions.base import Extension


class HotReloadAdminExtension(Extension):

    @staticmethod
    def reload_aimls(client_context):
        YLogger.debug(client_context, "Hot reloading all AIML files")
        client_context.brain.reload_aimls()
        return 'HOTRELOAD OK'

    @staticmethod
    def reload_denormal(client_context):
        YLogger.debug(client_context, "Hot reloading Denormal")
        client_context.brain.load_denormals()
        return 'HOTRELOAD OK'

    @staticmethod
    def reload_normal(client_context):
        YLogger.debug(client_context, "Hot reloading Normal")
        client_context.brain.load_normals()
        return 'HOTRELOAD OK'

    @staticmethod
    def reload_gender(client_context):
        YLogger.debug(client_context, "Hot reloading Gender")
        client_context.brain.load_genders()
        return 'HOTRELOAD OK'

    @staticmethod
    def reload_person(client_context):
        YLogger.debug(client_context, "Hot reloading Person")
        client_context.brain.load_persons()
        return 'HOTRELOAD OK'

    @staticmethod
    def reload_person2(client_context):
        YLogger.debug(client_context, "Hot reloading Person2")
        client_context.brain.load_person2s()
        return 'HOTRELOAD OK'

    @staticmethod
    def reload_properties(client_context):
        YLogger.debug(client_context, "Hot reloading Properties")
        client_context.brain.load_properties()
        return 'HOTRELOAD OK'

    @staticmethod
    def reload_defaults(client_context):
        YLogger.debug(client_context, "Hot reloading Defaults")
        client_context.brain.load_default_variables()
        return 'HOTRELOAD OK'

    @staticmethod
    def reload_regex(client_context):
        YLogger.debug(client_context, "Hot reloading Regex")
        client_context.brain.load_regex_templates()
        return 'HOTRELOAD OK'

    @staticmethod
    def reload_patterns(client_context):
        YLogger.debug(client_context, "Hot reloading Pattern Nodes")
        client_context.brain.aiml_parser.pattern_parser.pattern_factory.load(client_context.client.storage_factory)
        return 'HOTRELOAD OK'

    @staticmethod
    def reload_templates(client_context):
        YLogger.debug(client_context, "Hot reloading Template Nodes")
        client_context.brain.aiml_parser.template_parser.template_factory.load(client_context.client.storage_factory)
        return 'HOTRELOAD OK'

    @staticmethod
    def reload_maps(client_context):
        YLogger.debug(client_context, "Hot reloading Maps")
        client_context.brain.load_maps()
        return 'HOTRELOAD OK'

    @staticmethod
    def reload_map(client_context, mapname):
        YLogger.debug(client_context, "Hot reloading Map [%s]", mapname)
        client_context.brain.reload_map(mapname)
        return 'HOTRELOAD OK'

    @staticmethod
    def reload_sets(client_context):
        YLogger.debug(client_context, "Hot reloading Sets")
        client_context.brain.load_sets()
        return 'HOTRELOAD OK'

    @staticmethod
    def reload_set(client_context, setname):
        YLogger.debug(client_context, "Hot reloading Set [%s]", setname)
        client_context.brain.reload_set(setname)
        return 'HOTRELOAD OK'

    @staticmethod
    def reload_rdfs(client_context):
        YLogger.debug(client_context, "Hot reloading Rdfs")
        client_context.brain.load_rdfs()
        return 'HOTRELOAD OK'

    @staticmethod
    def reload_rdf(client_context, rdfname):
        YLogger.debug(client_context, "Hot reloading RDF [%s]", rdfname)
        client_context.brain.reload_rdf(rdfname)
        return 'HOTRELOAD OK'

    @staticmethod
    def reload_all(client_context):
        YLogger.debug(client_context, "Hot reloading all")
        HotReloadAdminExtension.reload_aimls(client_context)
        HotReloadAdminExtension.reload_denormal(client_context)
        HotReloadAdminExtension.reload_normal(client_context)
        HotReloadAdminExtension.reload_gender(client_context)
        HotReloadAdminExtension.reload_person(client_context)
        HotReloadAdminExtension.reload_person2(client_context)
        HotReloadAdminExtension.reload_properties(client_context)
        HotReloadAdminExtension.reload_defaults(client_context)
        HotReloadAdminExtension.reload_regex(client_context)
        HotReloadAdminExtension.reload_patterns(client_context)
        HotReloadAdminExtension.reload_templates(client_context)
        HotReloadAdminExtension.reload_sets(client_context)
        HotReloadAdminExtension.reload_maps(client_context)
        HotReloadAdminExtension.reload_rdfs(client_context)
        return "HOTRELOAD OK"

    # execute() is the interface that is called from the <extension> tag in the AIML
    def execute(self, client_context, data):
        YLogger.debug(client_context, "Hot Reload Admin - [%s]", data)

        try:
            commands = HotReloadAdminExtension.split_into_commands(data)

            if commands[0] == 'RELOAD':
                entity = commands[1]
                if entity in ['DENORMAL', 'NORMAL', 'GENDER', 'PERSON', 'PERSON2', 'PROPERTIES', 'DEFAULTS',
                              'REGEX', 'PATTERNS', 'TEMPLATES', 'SET', 'MAP', 'RDF']:
                    return self._reload_entity(entity, commands, client_context)

                elif entity == 'ALL':
                    return self._reload_all(commands, client_context)

                else:
                    return "Unknown RELOAD entity [%s]" % entity

            elif commands[0] == 'COMMANDS':
                return self._command()

            else:
                return self._unknown_command(commands[0])

        except Exception as e:
            YLogger.exception(client_context, "Failed to execute hot reload extension", e)

        return "Hot Reload Admin Error"

    def _reload_entity(self, entity, commands, client_context):

        if entity == 'DENORMAL':
            return HotReloadAdminExtension.reload_denormal(client_context)

        elif entity == 'NORMAL':
            return HotReloadAdminExtension.reload_normal(client_context)

        elif entity == 'GENDER':
            return HotReloadAdminExtension.reload_gender(client_context)

        elif entity == 'PERSON':
            return HotReloadAdminExtension.reload_person(client_context)

        elif entity == 'PERSON2':
            return HotReloadAdminExtension.reload_person2(client_context)

        elif entity == 'PROPERTIES':
            return HotReloadAdminExtension.reload_properties(client_context)

        elif entity == 'DEFAULTS':
            return HotReloadAdminExtension.reload_defaults(client_context)

        elif entity == 'REGEX':
            return HotReloadAdminExtension.reload_regex(client_context)

        elif entity == 'PATTERNS':
            return HotReloadAdminExtension.reload_patterns(client_context)

        elif entity == 'TEMPLATES':
            return HotReloadAdminExtension.reload_templates(client_context)

        elif entity == 'SET':
            if len(commands) == 3:
                setname = commands[2]
                return HotReloadAdminExtension.reload_set(client_context, setname)
            else:
                return "Missing Set name"

        elif entity == 'MAP':
            if len(commands) == 3:
                mapname = commands[2]
                return HotReloadAdminExtension.reload_map(client_context, mapname)
            else:
                return "Missing Map name"

        else:  # RDF
            if len(commands) == 3:
                rdfname = commands[2]
                return HotReloadAdminExtension.reload_rdf(client_context, rdfname)
            else:
                return "Missing RDF name"

    def _reload_all(self, commands, client_context):
        if len(commands) == 3:
            entities = commands[2]
            if entities == 'AIML':
                return HotReloadAdminExtension.reload_aimls(client_context)

            elif entities == 'MAPS':
                return HotReloadAdminExtension.reload_maps(client_context)

            elif entities == 'SETS':
                return HotReloadAdminExtension.reload_sets(client_context)

            elif entities == 'RDFS':
                return HotReloadAdminExtension.reload_rdfs(client_context)

            else:
                return "Unknown RELOAD ALL entity [%s]" % entities

        else:
            return HotReloadAdminExtension.reload_all(client_context)

    def _command(self):
        return "RELOAD [DENORMAL|NORMAL|GENDER|PERSON|PERSON2|PROPERTIES|DEFAULTS|REGEX|PATTERNS|TEMPLATES]" \
               " | [SET|MAP|RDF] NAME | ALL [AIML|MAPS|SETS|RDFS]"

    def _unknown_command(self, command):
        return "Unknown reload command [%s]" % command



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.storage.stores.sql.store.sqlstore import SQLStore
from programy.storage.entities.spelling import SpellingStore
from programy.storage.stores.sql.dao.corpus import Corpus
from programy.storage.entities.store import Store
from programy.utils.console.console import outputLog


class SQLSpellingStore(SQLStore, SpellingStore):

    def __init__(self, storage_engine):
        SQLStore.__init__(self, storage_engine)
        SpellingStore.__init__(self)

    def _get_all(self):
        return self._storage_engine.session.query(Corpus)

    def empty(self):
        return self._get_all()

    def _read_corpus_from_file(self, filename, verbose):
        count = 0
        success = 0
        with open(filename, "r") as text_file:
            for lines in text_file:
                words = lines.split(' ')
                for word in words:
                    corpus = Corpus(word=word)
                    self.storage_engine.session.add(corpus)
                    if verbose is True:
                        outputLog(self, word)
                    success += 1
                    count += 1

        return count, success

    def upload_from_file(self, filename, fileformat=Store.TEXT_FORMAT, commit=True, verbose=False):

        try:
            count, success = self._read_corpus_from_file(filename, verbose)
            self.commit(commit)
            return count, success

        except Exception as error:
            YLogger.exception(self, "Failed to load corpus from [%s]", error, filename)

        return 0, 0

    def load_spelling(self, spell_checker):
        corpus = self._storage_engine.session.query(Corpus)
        words = []
        for dbword in corpus:
            words.append(dbword.word)
        all_words = " ".join(words)
        spell_checker.add_corpus(all_words)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.config.base import BaseConfigurationData
from programy.storage.stores.nosql.mongo.engine import MongoStorageEngine
from programy.utils.substitutions.substitues import Substitutions


class MongoStorageConfiguration(BaseConfigurationData):

    def __init__(self):
        BaseConfigurationData.__init__(self, name="config")

        self._url = 'mongodb://localhost:27017/'
        self._database = "programy"
        self._drop_all_first = False

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, url):
        self._url = url

    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, database):
        self._database = database

    @property
    def drop_all_first(self):
        return self._drop_all_first

    @drop_all_first.setter
    def drop_all_first(self, drop_all):
        self._drop_all_first = drop_all

    def create_mongostorage_config(self):
        return {'url': self._url,
                'database': self._database,
                'drop_all_first': self._drop_all_first}

    def load_config_section(self, configuration_file, configuration, bot_root, subs: Substitutions = None):
        del bot_root
        del subs
        storage = configuration_file.get_section(self._section_name, configuration)
        if storage is not None:
            self._url = configuration_file.get_option(storage, "url")
            self._database = configuration_file.get_option(storage, "database")
            self._drop_all_first = configuration_file.get_option(storage, "drop_all_first")
        else:
            YLogger.error(None, "'config' section missing from storage config")

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data['url'] = 'mongodb://localhost:27017/'
            data['database'] = "programy"
            data['drop_all_first'] = True
        else:
            data['url'] = self._url
            data['database'] = self._database
            data['drop_all_first'] = self._drop_all_first

    def create_engine(self):
        engine = MongoStorageEngine(self)
        engine.initialise()
        return engine



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import nltk
from nltk.stem.regexp import RegexpStemmer
from nltk.stem.lancaster import LancasterStemmer
from nltk.stem.isri import ISRIStemmer
from nltk.stem.porter import PorterStemmer
from nltk.stem.snowball import SnowballStemmer
from nltk.stem.rslp import RSLPStemmer
from nltk.stem.cistem import Cistem


class Stemmer:

    @staticmethod
    def download_additional():
        nltk.download('rslp')   # pragma: no cover

    def __init__(self, stemmer="porter", **kwargs):
        self._impl = self._get_stemmer(stemmer, **kwargs)

    def stem(self, string):
        if self._impl is not None:
            return self._impl.stem(string)

        return string

    def _get_porter_stemmer(self):
        return PorterStemmer()

    def _get_lancaster_stemmer(self):
        return LancasterStemmer()

    def _get_regex_stemmer(self, regexp, minimum):
        return RegexpStemmer(regexp=regexp, min=minimum)

    def _get_iris_stemmer(self):
        return ISRIStemmer()

    def _get_snowball_stemmer(self, language):
        return SnowballStemmer(language=language)

    def _get_rslp_stemmer(self):
        return RSLPStemmer()

    def _get_cis_stemmer(self, case_insensitive):
        return Cistem(case_insensitive=case_insensitive)

    def _get_stemmer(self, stemmer="porter", **kwargs):

        if stemmer == "porter":
            return self._get_porter_stemmer()

        elif stemmer == "lancaster":
            return self._get_lancaster_stemmer()

        elif stemmer == "regex":
            regexp = kwargs['regexp']
            if 'min' in kwargs:
                minimum = kwargs['min']

            else:
                minimum = 0

            return self._get_regex_stemmer(regexp=regexp, minimum=minimum)

        elif stemmer == "isri":
            return self._get_iris_stemmer()

        elif stemmer == "snowball":
            if 'language' in kwargs:
                language = kwargs['language']

            else:
                language = 'english'

            return self._get_snowball_stemmer(language=language)

        elif stemmer == "rslp":
            return self._get_rslp_stemmer()

        elif stemmer == "cistem":
            if 'case_insensitive' in kwargs:
                case_insensitive = kwargs['case_insensitive']

            else:
                case_insensitive = False

            return self._get_cis_stemmer(case_insensitive=case_insensitive)

        else:
            raise ValueError("Unknown stemmer [%s]"%stemmer)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
import sys
import json
import yaml
from programy.clients.events.console.client import ConsoleBotClient
from programy.config.file.yaml_file import YamlConfigurationFile
from programy.config.programy import ProgramyConfiguration
from programy.clients.args import CommandLineClientArguments
from programy.utils.substitutions.substitues import Substitutions
from programy.clients.events.console.config import ConsoleConfiguration


class EmbeddedConfigFileBot(ConsoleBotClient):

    def __init__(self, config_filename, logging_filename=None):
        self._config_filename = config_filename

        filepath = os.path.dirname(__file__) + os.sep
        if logging_filename is not None:
            self._logging_filename = logging_filename
        else:
            self._logging_filename = filepath + 'basicbot/logging.yaml'

        ConsoleBotClient.__init__(self, "Console")

    def _render_callback(self):
        return False

    def parse_arguments(self, argument_parser):
        client_args = CommandLineClientArguments(self, parser=None)
        if self._logging_filename is not None:
            client_args._logging = self._logging_filename
        return client_args

    def load_configuration(self, arguments, subs: Substitutions = None):
        client_config = self.get_client_configuration()
        self._configuration = ProgramyConfiguration(client_config)

        yaml_file = YamlConfigurationFile()
        yaml_file.load_from_file(self._config_filename, client_config, ".")

    @staticmethod
    def generate_default_config_file(format):
        config = ConsoleConfiguration()
        data = {config.section_name: {}}
        config.to_yaml(data[config.section_name], True)
        if format == 'json':
            print(json.dumps(data, indent=4))

        elif format == 'yaml':
            print(yaml.dump(data, sort_keys=False))

        else:
            print ("Invalid data format 'json' or 'yaml' only!")

if __name__ == '__main__':

    if sys.argv[1] == 'run':
        config_file = sys.argv[2]
        logging_file = None
        if len(sys.argv) == 4:
            logging_file = sys.argv[3]

        print("Loading Bot Brain....please wait!")
        my_bot = EmbeddedConfigFileBot(config_file, logging_file)

        client_context = my_bot.create_client_context("testuser")
        response = my_bot.process_question(client_context, "Hello")
        print("Asked 'Hello', Response '%s'" % response)

    elif sys.argv[1] == 'config':
        EmbeddedConfigFileBot.generate_default_config_file(sys.argv[2])




"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from telegram.ext import Updater
from telegram.ext import CommandHandler
from telegram.ext import MessageHandler, Filters
from programy.utils.logging.ylogger import YLogger
from programy.clients.polling.client import PollingBotClient
from programy.clients.polling.telegram.config import TelegramConfiguration
from programy.utils.console.console import outputLog


def start(telegram_bot, update):
    if TelegramBotClient.TELEGRAM_CLIENT is None:
        raise Exception("Please initialise Telegram Client first")
    TelegramBotClient.TELEGRAM_CLIENT.start(telegram_bot, update)


def message(telegram_bot, update):
    if TelegramBotClient.TELEGRAM_CLIENT is None:
        raise Exception("Please initialise Telegram Client first")
    TelegramBotClient.TELEGRAM_CLIENT.message(telegram_bot, update)


def unknown(telegram_bot, update):
    if TelegramBotClient.TELEGRAM_CLIENT is None:
        raise Exception("Please initialise Telegram Client first")
    TelegramBotClient.TELEGRAM_CLIENT.unknown(telegram_bot, update)


class TelegramBotClient(PollingBotClient):

    TELEGRAM_CLIENT = None

    def __init__(self, argument_parser=None):
        self._updater = None
        self._telegram_token = None
        PollingBotClient.__init__(self, "telegram", argument_parser)

    def get_client_configuration(self):
        return TelegramConfiguration()

    def get_license_keys(self):
        self._telegram_token = self.license_keys.get_key("TELEGRAM_TOKEN")

    def create_updater(self, telegram_token):
        self._updater = Updater(token=telegram_token)

    def register_handlers(self):
        start_handler = CommandHandler('start', start)
        message_handler = MessageHandler(Filters.text, message)
        unknown_handler = MessageHandler(Filters.command, unknown)

        self._updater.dispatcher.add_handler(start_handler)
        self._updater.dispatcher.add_handler(message_handler)
        # Add unknown last
        self._updater.dispatcher.add_handler(unknown_handler)

    def get_initial_question(self, update):
        client_context = self.create_client_context(update.message.chat_id)
        initial_question = client_context.bot.get_initial_question(client_context)
        processed_question = client_context.bot.post_process_response(client_context,
                                                                      initial_question,
                                                                      srai=False)
        return processed_question

    def ask_question(self, userid, question):
        self._questions += 1
        client_context = self.create_client_context(userid)
        response = client_context.bot.ask_question(client_context, question, responselogger=self)
        return self.renderer.render(client_context, response)

    def start(self, telegram_bot, update):
        try:
            initial_question = self.get_initial_question(update)
            if initial_question:
                telegram_bot.send_message(chat_id=update.message.chat_id, text=initial_question)
            else:
                YLogger.error(self, "Not initial question to return in start()")

        except Exception as e:
            YLogger.exception(self, "Failed to start", e)

    def message(self, telegram_bot, update):
        try:
            response = self.ask_question(update.message.chat_id, update.message.text)
            if response:
                telegram_bot.send_message(chat_id=update.message.chat_id, text=response)
            else:
                YLogger.error(self, "Not response to return in message()")

        except Exception as e:
            YLogger.exception(self, "Failed to handle message", e)

    def get_unknown_response(self, userid):
        return self.ask_question(userid, self.configuration.client_configuration.unknown_command_srai)

    def get_unknown_command(self, userid):
        if self.configuration.client_configuration.unknown_command_srai is None:
            unknown_response = self.configuration.client_configuration.unknown_command
        else:
            unknown_response = self.get_unknown_response(userid)
            if unknown_response is None or unknown_response == "":
                unknown_response = self.configuration.client_configuration.unknown_command
        return unknown_response

    def unknown(self, telegram_bot, update):
        try:
            unknown_response = self.get_unknown_command(update.message.chat_id)
            if unknown_response:
                telegram_bot.send_message(chat_id=update.message.chat_id, text=unknown_response)
                YLogger.error(self, "No response to return in unknown()")

        except Exception as e:
            YLogger.exception(self, "Failed to handle unknown", e)

    def display_connected_message(self):
        outputLog(self, "Telegram Bot connected and running...")

    def connect(self):
        if self._telegram_token is not None:
            self.create_updater(self._telegram_token)
            self.register_handlers()
            return True

        outputLog(self, "No telegram token defined, unable to connect")
        return False

    def poll_and_answer(self):

        running = True
        try:
            self._updater.start_polling()
            # Without this the system goes into 100% CPU utilisation
            self._updater.idle()

        except KeyboardInterrupt:
            outputLog(self, "Telegram client stopping....")
            running = False
            self._updater.stop()

        except Exception as excep:
            YLogger.exception(self, "Failed to poll and answer", excep)

        return running


if __name__ == '__main__':

    outputLog(None, "Initiating Telegram Client...")

    TelegramBotClient.TELEGRAM_CLIENT = TelegramBotClient()
    TelegramBotClient.TELEGRAM_CLIENT.run()



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from programy.utils.logging.ylogger import YLogger

from programy.config.base import BaseConfigurationData
from programy.utils.substitutions.substitues import Substitutions


class EmailConfiguration(BaseConfigurationData):

    def __init__(self):
        BaseConfigurationData.__init__(self, name="email")
        self._host = None
        self._port = None
        self._username = None
        self._password = None
        self._from_addr = None

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def username(self):
        return self._username

    @property
    def password(self):
        return self._password

    @property
    def from_addr(self):
        return self._from_addr

    def check_for_license_keys(self, license_keys):
        username = self._extract_license_key(self._username, license_keys)
        if username is not None:
            self._username = username

        password = self._extract_license_key(self._password, license_keys)
        if password is not None:
            self._password = password

        super(EmailConfiguration, self).check_for_license_keys(license_keys)

    def load_config_section(self, configuration_file, configuration, bot_root, subs: Substitutions = None):
        del bot_root
        del subs
        email = configuration_file.get_section(self._section_name, configuration)
        if email is not None:
            self._host = configuration_file.get_option(email, "host", missing_value=None)
            self._port = configuration_file.get_option(email, "port", missing_value=None)
            self._username = configuration_file.get_option(email, "username", missing_value=None)
            self._password = configuration_file.get_option(email, "password", missing_value=None)
            self._from_addr = configuration_file.get_option(email, "from_addr", missing_value=None)
            if self._from_addr is None:
                self._from_addr = self._username
        else:
            YLogger.warning(self, "'email' section missing from bot config, using defaults")

    def to_yaml(self, data, defaults=True):
        if defaults is True:
            data['host'] = None
            data['port'] = None
            data['username'] = None
            data['password'] = None
            data['from_addr'] = None
        else:
            data['host'] = self._host
            data['port'] = self._port
            data['username'] = self._username
            data['password'] = self._password
            data['from_addr'] = self._from_addr



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.parser.template.nodes.base import TemplateNode
from programy.parser.exceptions import ParserException


class TemplateAuthoriseNode(TemplateNode):

    def __init__(self):
        TemplateNode.__init__(self)
        self._role = None
        self._denied_srai = None

    @property
    def role(self):
        return self._role

    @role.setter
    def role(self, role):
        self._role = role

    @property
    def denied_srai(self):
        return self._denied_srai

    @denied_srai.setter
    def denied_srai(self, denied_srai):
        self._denied_srai = denied_srai

    def resolve_to_string(self, client_context):
        # Check if the user, role or group exists, assumption being, that if defined
        # in the tag and exists then we can execute the inner children
        # Assumption is that user has been authenticated and passed and is value
        if client_context.brain.security.authorisation is not None:
            if client_context.brain.security.authorisation.authorise(client_context.userid, self.role) is False:
                if self._denied_srai is not None:
                    srai_text = self._denied_srai
                else:
                    srai_text = client_context.brain.security.authorisation.get_default_denied_srai()
                resolved = client_context.bot.ask_question(client_context, srai_text, srai=True)
                YLogger.debug(self, "[%s] resolved to [%s]", self.to_string(), resolved)
                return resolved

        # Resolve afterwards, as pointless resolving before checking for authorisation
        resolved = self.resolve_children_to_string(client_context)
        YLogger.debug(self, "[%s] resolved to [%s]", self.to_string(), resolved)
        return resolved

    def to_string(self):
        text = "[AUTHORISE ("
        text += "role=%s" % self._role
        if self._denied_srai is not None:
            text += ", denied_srai=%s" % self._denied_srai
        text += ")]"
        return text

    def to_xml(self, client_context):
        xml = '<authorise'
        xml += ' role="%s"' % self._role
        if self._denied_srai is not None:
            xml += ' denied_srai="%s"' % self._denied_srai
        xml += '>'
        xml += self.children_to_xml(client_context)
        xml += '</authorise>'
        return xml

    #######################################################################################################
    # AUTHORISE_ATTRIBUTES ::= role="ROLEID" [, denied_srai="SRAI_TAG"]
    # AUTHORISE_EXPRESSION ::== <authorise( AUTHORISE_ATTRIBUTES)*>TEMPLATE_EXPRESSION</authorise> |

    def parse_expression(self, graph, expression):

        if 'role' in expression.attrib:
            self._role = expression.attrib['role']

        if self._role is None:
            raise ParserException("AUTHORISE role attribute missing !")

        if self._role == "":
            raise ParserException("AUTHORISE role attribute empty !")

        if 'denied_srai' in expression.attrib:
            self._denied_srai = expression.attrib['denied_srai']

            if self._denied_srai == "":
                raise ParserException("AUTHORISE denied_srai attribute empty !")

        head_text = self.get_text_from_element(expression)
        self.parse_text(graph, head_text)

        for child in expression:
            graph.parse_tag_expression(child, self)
            tail_text = self.get_tail_from_element(child)
            self.parse_text(graph, tail_text)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.console.console import outputLog


class OpenchatBotReponseObject:

    @staticmethod
    def get_parameter(data, name, defaultValue=None):
        if name in data:
            return data[name]
        else:
            return defaultValue


class OpenChatBotMediaDefaultAction(OpenchatBotReponseObject):

    def __init__(self, actiontype, label, payload):
        self.actiontype = actiontype
        self.label = label
        self.payload = payload

    @staticmethod
    def parse(json_data):
        actiontype = OpenchatBotReponseObject.get_parameter(json_data, 'type', None)
        label = OpenchatBotReponseObject.get_parameter(json_data, 'label', None)
        payload = OpenchatBotReponseObject.get_parameter(json_data, 'payload', None)
        return OpenChatBotMediaDefaultAction(actiontype, label, payload)

    def to_aiml(self):
        if self.actiontype == 'web_url':
            aiml = "<link>"
            aiml += "<text>%s</text>" % self.label
            if self.payload is not None:
                aiml += "<url>%s</url>" % self.payload
            aiml += "</link>"
            return aiml

        outputLog(None, "Unknown Default Action type [%s]" % self.actiontype)
        return None


class OpenChatBotMediaButton(OpenchatBotReponseObject):

    def __init__(self, actiontype, client, label, payload):
        self.actiontype = actiontype
        self.client = client
        self.label = label
        self.payload = payload

    @staticmethod
    def parse(json_data):
        actiontype = OpenchatBotReponseObject.get_parameter(json_data, 'type', None)
        client = OpenchatBotReponseObject.get_parameter(json_data, 'client', None)
        label = OpenchatBotReponseObject.get_parameter(json_data, 'label', None)
        payload = OpenchatBotReponseObject.get_parameter(json_data, 'payload', None)
        return OpenChatBotMediaButton(actiontype, client, label, payload)

    def to_aiml(self):
        if self.actiontype == 'web_url':
            aiml = "<button>"
            aiml += "<text>%s</text>" % self.label
            if self.payload is not None:
                aiml += "<url>%s</url>" % self.payload
            aiml += "</button>"
            return aiml

        outputLog(self, "Unknown Button type [%s]" % self.actiontype)
        return None


class OpenChatBotMedia(OpenchatBotReponseObject):

    def __init__(self, shortDesc, longDesc, title, mimeType, src, default_action, buttons):
        self.shortDesc = shortDesc
        self.longDesc = longDesc
        self.title = title
        self.mimeType = mimeType
        self.src = src
        self.default_action = default_action
        self.buttons = buttons

    @staticmethod
    def parse(json_data):
        shortDesc = OpenchatBotReponseObject.get_parameter(json_data, 'shortDesc', None)
        longDesc = OpenchatBotReponseObject.get_parameter(json_data, 'longDesc', None)
        title = OpenchatBotReponseObject.get_parameter(json_data, 'title', None)
        mimeType = OpenchatBotReponseObject.get_parameter(json_data, 'mimeType', None)
        src = OpenchatBotReponseObject.get_parameter(json_data, 'src', None)

        default_action = None
        if 'default_action' in json_data:
            default_action = OpenChatBotMediaDefaultAction.parse(json_data['default_action'])

        buttons = None
        if 'buttons' in json_data:
            buttons = [OpenChatBotMediaButton.parse(button) for button in json_data['buttons']]

        return OpenChatBotMedia(shortDesc, longDesc, title, mimeType, src, default_action, buttons)

    def to_aiml(self):
        aiml = "<card>"
        aiml += "<image>%s</image>" % self.src
        aiml += "<title>%s</title>" % self.shortDesc
        aiml += "<subtitle>%s</subtitle>" % self.longDesc
        for button in self.buttons:
            if button.actiontype == 'web_url':
                aiml += "<button>"
                aiml += "<text>%s</text>" % button.label
                if button.payload is not None:
                    aiml += "<url>%s</url>" % button.payload
                aiml += "</button>"
        aiml += "</card>"
        return aiml


class OpenChatBotTTS(OpenchatBotReponseObject):

    def __init__(self, actiontype, payload):
        self.actiontype = actiontype
        self.payload = payload

    @staticmethod
    def parse(json_data):
        actiontype = OpenchatBotReponseObject.get_parameter(json_data, 'type', None)
        payload = OpenchatBotReponseObject.get_parameter(json_data, 'payload', None)
        return OpenChatBotTTS(actiontype, payload)

    def to_aiml(self):
        aiml = "<tts>"
        aiml += "<type>%s</type>" % self.actiontype
        aiml += self.payload
        aiml += "</tts>"
        return aiml


class OpenChatBotSuggestion(OpenchatBotReponseObject):

    def __init__(self, actiontype, label, payload):
        self.actiontype = actiontype
        self.label = label
        self.payload = payload

    @staticmethod
    def parse(json_data):
        actiontype = OpenchatBotReponseObject.get_parameter(json_data, 'type', None)
        label = OpenchatBotReponseObject.get_parameter(json_data, 'label', None)
        payload = OpenchatBotReponseObject.get_parameter(json_data, 'payload', None)
        return OpenChatBotSuggestion(actiontype, label, payload)

    def to_aiml(self):
        if self.actiontype == 'web_url':
            aiml = "<link>"
            aiml += "<text>%s</text>" % self.label
            if self.payload is not None:
                aiml += "<url>%s</url>" % self.payload
            aiml += "</link>"
            return aiml

        outputLog(None, "Unknown Suggestion type [%s]" % self.actiontype)
        return None


class OpenChatBotResponse(OpenchatBotReponseObject):

    def __init__(self, query, userId, timestamp, text, tts, infoURL, media, suggestions, context):
        self.query = query
        self.userId = userId
        self.timestamp = timestamp
        self.text = text
        self.tts = tts
        self.infoURL = infoURL
        self.media = media
        self.suggestions = suggestions
        self.context = context

    @staticmethod
    def parse(json_data):
        query = OpenchatBotReponseObject.get_parameter(json_data, 'query', None)
        userId = OpenchatBotReponseObject.get_parameter(json_data, 'userId', None)
        timestamp = OpenchatBotReponseObject.get_parameter(json_data, 'timestamp', None)
        text = OpenchatBotReponseObject.get_parameter(json_data, 'text', None)
        infoURL = OpenchatBotReponseObject.get_parameter(json_data, 'infoURL', None)

        tts = None
        if 'tts' in json_data:
            tts = OpenChatBotTTS.parse(json_data['tts'])

        media = None
        if 'media' in json_data:
            media = [OpenChatBotMedia.parse(mediax) for mediax in json_data['media']]

        suggestions = None
        if 'suggestions' in json_data:
            suggestions = [OpenChatBotSuggestion.parse(suggestionx) for suggestionx in json_data['suggestions']]

        contexts = []
        if 'context' in json_data:
            contexts = json_data['context'][:]

        return OpenChatBotResponse(query, userId, timestamp, text, tts, infoURL, media, suggestions, contexts)

    def to_aiml(self):
        aiml = ""

        if self.text is not None:
            aiml += self.text

        if self.tts is not None:
            aiml += " "
            aiml += self.tts.to_aiml()

        if self.media:
            for media in self.media:
                media_aiml = media.to_aiml()
                if media_aiml:
                    aiml += " "
                    aiml += media_aiml

        if self.suggestions:
            for suggestion in self.suggestions:
                sugg_aiml = suggestion.to_aiml()
                if sugg_aiml:
                    aiml += " "
                    aiml += sugg_aiml

        return aiml


class OpenChatBotStatus(OpenchatBotReponseObject):

    def __init__(self, code, message):
        self.code = code
        self.message = message

    @staticmethod
    def parse(json_data):
        code = OpenchatBotReponseObject.get_parameter(json_data, 'code', None)
        if code is None:
            return None
        message = OpenchatBotReponseObject.get_parameter(json_data, 'message', None)
        return OpenChatBotStatus(code, message)


class OpenChatBotMeta(OpenchatBotReponseObject):

    def __init__(self, botName, botIcon, version, copyr, authors):
        self.botName = botName
        self.botIcon = botIcon
        self.version = version
        self.copyr = copyr
        self.authors = authors

    @staticmethod
    def parse(json_data):
        botName = OpenchatBotReponseObject.get_parameter(json_data, 'botName', None)
        botIcon = OpenchatBotReponseObject.get_parameter(json_data, 'botIcon', None)
        version = OpenchatBotReponseObject.get_parameter(json_data, 'version', None)
        copyr = OpenchatBotReponseObject.get_parameter(json_data, 'copyright', None)
        authors = OpenchatBotReponseObject.get_parameter(json_data, 'authors', None)
        return OpenChatBotMeta(botName, botIcon, version, copyr, authors)



import sys
import os.path
import xml.etree.ElementTree as ET


if __name__ == '__main__':

    aiml_dir = sys.argv[1]
    csv_file = sys.argv[2]

    print("aiml_dir:", aiml_dir)
    print("csv_file:", csv_file)

    questions = []

    files = 0
    for dirpath, dirnames, filenames in os.walk(aiml_dir):
        for filename in filenames:
            files += 1
            aiml_file = os.path.join(dirpath, filename)
            print (aiml_file)

            try:
                tree = ET.parse(aiml_file)
                aiml = tree.getroot()
                categories = aiml.findall('category')
                for category in categories:
                    pattern_text = ""

                    pattern = category.find("pattern")
                    for elt in pattern.iter():

                        comma = False
                        if elt.tag == "pattern":
                            if elt.text is not None:
                                text = elt.text.strip().upper()
                                pattern_text += " ".join(text.split())
                                comma = True

                        elif elt.tag == "set":
                            if 'name' in elt.attrib:
                                name = elt.attrib['name']
                            else:
                                name = elt.text.strip()
                            if comma is True:
                                pattern_text += " "
                            pattern_text += " SET[%s]"%name
                            if text:
                                pattern_text += " "
                                pattern_text += " ".join(text.split())
                            comma = True

                        elif elt.tag == "bot":
                            if 'name' in elt.attrib:
                                name = elt.attrib['name']
                            else:
                                name = elt.text.strip()
                            if comma is True:
                                pattern_text += " "
                            pattern_text += " BOT[%s]" % name
                            if text:
                                pattern_text += " "
                                pattern_text += " ".join(text.split())
                            comma = True

                        if elt.tail is not None and elt.tail.strip() != "":
                            if comma is True:
                                pattern_text += " "
                            text = elt.tail.strip().upper()
                            if text:
                                pattern_text += " "
                                pattern_text += " ".join(text.split())
                            comma = True

                        if pattern_text is not None:
                            pattern_text = pattern_text.strip()
                            if len(pattern_text) > 0:
                                questions.append([aiml_file, pattern_text])

            except Exception as e:
                print(e)
                raise e

    questions.sort(key=lambda x: x[1])
    with open(csv_file, "w+") as output_file:
        for line in questions:
            new_line = ", ".join(line[1].split())
            output_file.write(line[0])
            output_file.write(", ")
            output_file.write(new_line)
            output_file.write("\n")

    print("Files:    %d"%files)
    print("Patterns: %d"%len(questions))


"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger

from programy.config.brain.security import BrainSecurityConfiguration
from programy.security.authenticate.authenticator import Authenticator


class ClientIdAuthenticationService(Authenticator):

    def __init__(self, configuration: BrainSecurityConfiguration):
        Authenticator.__init__(self, configuration)
        self.authorised = [
            "console"
        ]

    def user_auth_service(self, client_context):
        del client_context
        return False

    # Its at this point that we would call a user auth service, and if that passes
    # return true, appending the user to the known authorised list of user
    # This is a very naive approach, and does not cater for users that log out, invalidate
    # their credentials, or have a TTL on their credentials
    # #Exercise for the reader......
    def _auth_clientid(self, client_context):
        authorised = self.user_auth_service(client_context)
        if authorised is True:
            self.authorised.append(client_context.userid)
        return authorised

    def authenticate(self, client_context):
        try:
            if client_context.userid in self.authorised:
                return True
            else:
                if self._auth_clientid(client_context) is True:
                    return True

                return False
        except Exception as excep:
            YLogger.error(client_context, str(excep))
            return False



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
import os.path
from programy.storage.utils.processors import TextFile
from programy.storage.utils.processors import CSVFileReader
from programy.utils.logging.ylogger import YLogger


class Store:
    TEXT_FORMAT = "text"
    CSV_FORMAT = "csv"
    XML_FORMAT = "xml"
    BINARY_FORMAT = "bin"
    YAML_FORMAT = "yaml"

    def __init__(self):
        pass                # pragma: no cover

    def empty(self):
        return              # pragma: no cover

    def empty_named(self, name):
        del name            # pragma: no cover
        return              # pragma: no cover

    def commit(self, commit=True):
        return              # pragma: no cover

    def rollback(self, commit=True):
        return              # pragma: no cover

    def get_split_char(self):
        return ","

    def split_into_fields(self, line):
        return line.split(",")

    def process_line(self, name, fields, verbose=False):
        del name            # pragma: no cover
        del fields          # pragma: no cover
        del verbose         # pragma: no cover
        return False

    def upload_from_text(self, name, text, commit=True):
        try:
            lines = text.split('\n')
            for line in lines:
                line = line.strip()
                if line and len(line)>0:
                    fields = self.split_into_fields(line)
                    self.process_line(name, fields)

            self.commit(commit)

        except Exception as e:
            YLogger.exception_nostack(self, "Error loading from text", e)
            self.rollback(commit)

    @staticmethod
    def get_file_processor(fileformat, filename):
        if fileformat == Store.TEXT_FORMAT:
            return TextFile(filename)
        elif fileformat == Store.CSV_FORMAT:
            return CSVFileReader(filename)
        else:
            raise Exception("Unknown file format [%s]" % fileformat)

    @staticmethod
    def get_just_filename_from_filepath(filepath):

        if os.sep in filepath:
            pathsplits = filepath.split(os.sep)
            filename_ext = pathsplits[-1]
        else:
            filename_ext = filepath

        if "." in filename_ext:
            filesplits = filename_ext.split(".")
            filename = filesplits[0]
        else:
            filename = filename_ext

        return filename.upper()

    def upload_from_directory(self, directory, fileformat=TEXT_FORMAT, extension=None, subdir=True, commit=True,
                              verbose=False):

        final_count = 0
        final_success = 0

        try:
            if subdir is False:
                paths = os.listdir(directory)
                for filename in paths:
                    fullpath = os.path.join(directory, filename)
                    if os.path.isdir(fullpath) is False:
                        if extension is not None:
                            if filename.endswith(extension):
                                count, success = self.upload_from_file(fullpath, fileformat=fileformat, commit=commit,
                                                                       verbose=verbose)
                                final_count += count
                                final_success += success
                        else:
                            count, success = self.upload_from_file(fullpath, fileformat=fileformat, commit=commit,
                                                                   verbose=verbose)
                            final_count += count
                            final_success += success
            else:
                for dirpath, _, filenames in os.walk(directory):
                    for filename in filenames:
                        if extension is not None:
                            if filename.endswith(extension):
                                count, success = self.upload_from_file(os.path.join(dirpath, filename),
                                                                       fileformat=fileformat, commit=commit,
                                                                       verbose=verbose)
                                final_count += count
                                final_success += success
                        else:
                            count, success = self.upload_from_file(os.path.join(dirpath, filename),
                                                                   fileformat=fileformat, commit=commit,
                                                                   verbose=verbose)
                            final_count += count
                            final_success += success

            self.commit(commit)

        except Exception as e:
            YLogger.exception_nostack(self, "Error loading from directory", e)
            self.rollback(commit)

        return final_count, final_success

    def upload_from_file(self, filename, fileformat=TEXT_FORMAT, commit=True, verbose=False):

        file_processor = None
        final_count = 0
        final_success = 0
        try:
            name = Store.get_just_filename_from_filepath(filename)

            file_processor = Store.get_file_processor(fileformat, filename)
            count, success = file_processor.process_lines(name, self, verbose=verbose)
            final_count += count
            final_success += success

            self.commit(commit)

        except Exception as e:
            YLogger.exception_nostack(self, "Error uploading from file", e)
            self.rollback(commit=commit)

        finally:
            if file_processor is not None:
                file_processor.close()

        return final_count, final_success



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


class OpenChatBot:

    def __init__(self, name, url, methods, authorization=None, api_key=None):
        self._name = name
        self._url = url
        self._methods = methods
        self._authorization = authorization
        self._api_key = api_key

    @property
    def name(self):
        return self._name

    @property
    def url(self):
        return self._url

    @property
    def methods(self):
        return self._methods

    @property
    def authorization(self):
        return self._authorization

    @property
    def api_key(self):
        return self._api_key

    @staticmethod
    def uri(host, port, endpoint):
        if endpoint.startswith('/') is False:
            endpoint = "/" + endpoint

        if port == 0:
            return "%s%s" % (host, endpoint)
        else:
            return "%s:%d%s" % (host, port, endpoint)

    @staticmethod
    def create(name, payload):

        if 'openchatbot' not in payload:
            return None
        openchatbot = payload['openchatbot']

        if 'host' not in openchatbot:
            return None
        host = openchatbot['host']

        if 'port' not in openchatbot:
            port = 0
        else:
            port = openchatbot['port']

        if 'endpoint' not in openchatbot:
            return None
        endpoint = openchatbot['endpoint']

        uri = OpenChatBot.uri(host, port, endpoint)

        if 'methods' not in openchatbot:
            return None
        methods = openchatbot['methods']

        return OpenChatBot(name, uri, methods)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import uuid
import datetime
from flask import Flask
from flask import jsonify
from flask import request
from flask import make_response
from flask import abort
from flask import current_app
from programy.utils.logging.ylogger import YLogger
from programy.clients.restful.flask.client import FlaskRestBotClient
from programy.clients.restful.flask.webchat.config import WebChatConfiguration
from programy.clients.render.html import HtmlRenderer
from programy.clients.restful.auth import RestAuthorizationHandler
from programy.utils.console.console import outputLog


class WebChatBotClient(FlaskRestBotClient):

    def __init__(self, argument_parser=None):
        FlaskRestBotClient.__init__(self, "WebChat", argument_parser)
        self._authorization = RestAuthorizationHandler.load_authorisation(self)

    def _render_callback(self):
        return False

    def get_client_configuration(self):
        return WebChatConfiguration()

    def get_default_renderer(self, callback=True):
        return HtmlRenderer()

    def unauthorised_access_response(self, error_code=401):
        return make_response(jsonify({'error': 'Unauthorized access'}), error_code)

    def get_question(self, request):
        if 'question' in request.args:
            return request.args['question']
        return None

    def get_userid(self, request):
        userid = request.cookies.get(self.configuration.client_configuration.cookie_id)
        if userid is None:
            userid = str(uuid.uuid4().hex)
            YLogger.debug(self, "Setting userid cookie to :%s" % userid)
        else:
            YLogger.debug(self, "Found userid cookie : %s" % userid)
        return userid

    def get_userid_cookie_expirary_date(self, duration):
        expire_date = datetime.datetime.now()
        expire_date = expire_date + datetime.timedelta(days=duration)
        return expire_date

    def create_success_response_data(self, question, answer):
        return {"question": question, "answer": answer}

    def get_default_response(self, client_context):
        return client_context.bot.default_response

    def create_error_response_data(self, client_context, question, error):
        return {"question": question,
                "answer": self.get_default_response(client_context),
                "error": error
                }

    def create_webchat_response(self, response_data, userid, userid_expire_date):
        response = jsonify({'response': response_data})
        response.set_cookie(self.configuration.client_configuration.cookie_id, userid, expires=userid_expire_date)
        return response

    def get_answer(self, client_context, question):
        if question == 'YINITIALQUESTION':
            answer = client_context.bot.get_initial_question(client_context)
        else:
            self._questions += 1
            answer = client_context.bot.ask_question(client_context, question, responselogger=self)
        return answer

    def receive_message(self, request):

        if self._authorization is not None:
            if self._authorization.authorise(request) is False:
                abort(403)

        if self._api_keys is not None:
            if self._api_keys.use_api_keys():
                if self._api_keys.verify_api_key_usage(request) is False:
                    abort(403)

        question = self.get_question(request)
        if question is None:
            YLogger.error(self, "'question' missing from request")
            abort(400)

        userid = self.get_userid(request)

        userid_expire_date = self.get_userid_cookie_expirary_date(self.configuration.client_configuration.cookie_expires)

        client_context = self.create_client_context(userid)
        try:
            answer = self.get_answer(client_context, question)
            answer = answer.replace('\n', '').strip()
            rendered = self._renderer.render(client_context, answer)
            response_data = self.create_success_response_data(question, rendered)

        except Exception as excep:
            YLogger.exception(self, "Failed receving message", excep)
            response_data = self.create_error_response_data(client_context, question, str(excep))

        return self.create_webchat_response(response_data, userid, userid_expire_date)


if __name__ == '__main__':

    outputLog(None, "Initiating WebChat Client...")

    APP = Flask(__name__)

    WEB_CLIENT = WebChatBotClient()

    @APP.route('/')
    def index():
        return current_app.send_static_file('webchat.html')

    @APP.route(WEB_CLIENT.configuration.client_configuration.api, methods=['GET'])
    def receive_message():
        try:
            return WEB_CLIENT.receive_message(request)
        except Exception as e:
            outputLog(None, "Error receiving webchat message", e)
            YLogger.exception(None, "Web client error", e)
            return "500"

    if WEB_CLIENT.ping_responder.config.url is not None:
        @APP.route(WEB_CLIENT.ping_responder.config.url, methods=['GET'])
        def ping():
            return jsonify(WEB_CLIENT.ping_responder.ping())

    if WEB_CLIENT.ping_responder.config.shutdown is not None:
        @APP.route(WEB_CLIENT.ping_responder.config.shutdown, methods=['GET'])
        def shutdown():
            WEB_CLIENT.ping_responder.stop_ping_service()
            return 'Server shutting down...'

    WEB_CLIENT.run(APP)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import http.server
import json
import asyncio
from botbuilder.schema import (Activity, ActivityTypes)
from botframework.connector import ConnectorClient
from botframework.connector.auth import (MicrosoftAppCredentials, JwtTokenValidation, SimpleCredentialProvider)
from programy.utils.logging.ylogger import YLogger
from programy.clients.restful.flask.client import FlaskRestBotClient
from programy.clients.restful.asyncio.microsoft.config import MicrosoftConfiguration
from programy.utils.console.console import outputLog


class MicrosoftBotClient(FlaskRestBotClient):

    def __init__(self, argument_parser=None):
        FlaskRestBotClient.__init__(self, 'microsoft', argument_parser)

        YLogger.debug(self, "Microsoft Client is running....")

        outputLog(None, "Microsoft Client loaded")

    def get_client_configuration(self):
        return MicrosoftConfiguration()

    def get_microsoft_app_id(self):
        return self.license_keys.get_key("MICROSOFT_APP_ID")

    def get_microsoft_app_password(self):
        return self.license_keys.get_key("MICROSOFT_APP_PASSWORD")

    def get_new_user_message(self):
        if self.configuration.client_configuration.new_user_srai is not None:
            pass

        return self.configuration.client_configuration.new_user_text

    def ask_question(self, userid, question, metadata=None):
        reply = ""
        try:
            client_context = self.create_client_context("microsoft")
            self._questions += 1
            response = client_context.bot.ask_question(client_context, question, responselogger=self)
            reply = self.renderer.render(client_context, response)

        except Exception as e:
            YLogger.exception(self, "Error getting reply from bot", e)

        return reply


MICROSOFT_CLIENT = MicrosoftBotClient()


class BotRequestHandler(http.server.BaseHTTPRequestHandler):

    @staticmethod
    def __create_reply_activity(request_activity, text):
        return Activity(
            type=ActivityTypes.message,
            channel_id=request_activity.channel_id,
            conversation=request_activity.conversation,
            recipient=request_activity.from_property,
            from_property=request_activity.recipient,
            text=text,
            service_url=request_activity.service_url)

    def __handle_conversation_update_activity(self, activity):
        self.send_response(202)
        self.end_headers()
        if len(activity.members_added):
            if activity.members_added[0].id != activity.recipient.id:
                credentials = MicrosoftAppCredentials(MICROSOFT_CLIENT.get_microsoft_app_id(),
                                                      MICROSOFT_CLIENT.get_microsoft_app_password())

                response = MICROSOFT_CLIENT.get_new_user_message()
                reply = BotRequestHandler.__create_reply_activity(activity, response)

                connector = ConnectorClient(credentials, base_url=reply.service_url)
                connector.conversations.send_to_conversation(reply.conversation.id, reply)

    def __handle_message_activity(self, activity):
        self.send_response(200)
        self.end_headers()
        credentials = MicrosoftAppCredentials(MICROSOFT_CLIENT.get_microsoft_app_id(),
                                              MICROSOFT_CLIENT.get_microsoft_app_password())
        connector = ConnectorClient(credentials, base_url=activity.service_url)

        response = MICROSOFT_CLIENT.ask_question(activity.recipient.id, activity.text)
        reply = BotRequestHandler.__create_reply_activity(activity, response)

        connector.conversations.send_to_conversation(reply.conversation.id, reply)

    def __handle_authentication(self, activity):
        credential_provider = SimpleCredentialProvider(MICROSOFT_CLIENT.get_microsoft_app_id(),
                                                       MICROSOFT_CLIENT.get_microsoft_app_password())
        loop = asyncio.new_event_loop()
        try:
            loop.run_until_complete(JwtTokenValidation.authenticate_request(activity,
                                                                            self.headers.get("Authorization"),
                                                                            credential_provider))
            return True
        except Exception as ex:
            self.send_response(401, ex)
            self.end_headers()
            return False
        finally:
            loop.close()

    def __unhandled_activity(self):
        self.send_response(404)
        self.end_headers()

    def do_POST(self):
        body = self.rfile.read(int(self.headers['Content-Length']))
        data = json.loads(str(body, 'utf-8'))
        activity = Activity.deserialize(data)

        if not self.__handle_authentication(activity):
            return

        if activity.type == ActivityTypes.conversation_update.value:
            self.__handle_conversation_update_activity(activity)
        elif activity.type == ActivityTypes.message.value:
            self.__handle_message_activity(activity)
        else:
            self.__unhandled_activity()


if __name__ == '__main__':

    outputLog(None, "Initiating Microsoft Client...")

    SERVER = None
    try:
        host = MICROSOFT_CLIENT.configuration.client_configuration.host
        port = MICROSOFT_CLIENT.configuration.client_configuration.port

        SERVER = http.server.HTTPServer((host, port), BotRequestHandler)
        outputLog(None, 'Started http server')
        SERVER.serve_forever()

    except KeyboardInterrupt:
        outputLog(None, 'Ctrl received, shutting down server')
        if SERVER is not None:
            SERVER.socket.close()



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import re
from slackclient import SlackClient
from programy.utils.logging.ylogger import YLogger
from programy.clients.polling.client import PollingBotClient
from programy.clients.polling.slack.config import SlackConfiguration
from programy.utils.console.console import outputLog


SLACK_CLIENT = None


class SlackBotClient(PollingBotClient):

    # constants
    DIRECT_REGEX = r"^<@(.+)>\s(.*)"
    MENTION_REGEX = r"^(.*)\s<@(|[WU].+?)>\s(.*)"

    def __init__(self, argument_parser=None):
        self._bot_token = None
        self._starterbot_id = None
        self._polling_interval = 1000

        PollingBotClient.__init__(self, "slack", argument_parser)

        self._slack_client = self.create_client()

    def get_client_configuration(self):
        return SlackConfiguration()

    def parse_configuration(self):
        self._polling_interval = self.configuration.client_configuration.polling_interval

    def get_license_keys(self):
        self._bot_token = self.license_keys.get_key("SLACK_TOKEN")

    def create_client(self):
        return SlackClient(self._bot_token)

    def connect(self):
        if self._slack_client.rtm_connect(with_team_state=False) is True:
            # Read bot's user ID by calling Web API method `auth.test`
            self._starterbot_id = self.get_bot_id()
            if self._starterbot_id is not None:
                return True

        return False

    def get_bot_id(self):
        return self._slack_client.api_call("auth.test")["user_id"]

    def parse_messages(self, slack_events):
        """
            Parses a list of events coming from the Slack RTM API to find bot messages.
            If a bot message is found, this function returns a tuple of message and channel.
            If its not found, then this function returns None, None.
        """
        for event in slack_events:
            if event["type"] == "message" and "subtype" not in event:
                self.parse_message(event)

    def parse_message(self, event):
        YLogger.debug(self, "Slack Bot Received: %s", event)

        text = event["text"]
        direct = False
        mention = False

        userid, message = self.parse_direct_message(text)
        if userid is None and message is None:
            userid, message = self.parse_mention(text)
            if userid is None and message is None:
                userid = event['user']
                message = text
            else:
                mention = True
        else:
            direct = True

        response = {
            "botid": self._starterbot_id,
            "userid": userid,
            "message": message
        }

        if message:
            channel = event['channel']

            if self.configuration.client_configuration.reply_to_all is True:
                response['reply_all'] = True
                self.handle_message(message, channel, userid)

            elif self.configuration.client_configuration.reply_to_direct is True and direct is True:
                response['direct'] = True
                self.handle_message(message, channel, userid)

            elif self.configuration.client_configuration.reply_to_mention is True and mention is True:
                response['mention'] = True
                self.handle_message(message, channel, userid)

            else:
                response['no_response'] = True

        YLogger.debug("Slack Bot Response: %s", response)


    def parse_mention(self, message_text):
        """
            Finds a direct mention (a mention that is at the beginning) in message text
            and returns the user ID which was mentioned. If there is no direct mention, returns None
        """
        matches = re.search(self.MENTION_REGEX, message_text)
        # the first group contains the username, the second group contains the remaining message
        if matches:
            return matches.group(2), matches.group(3).strip()

        return None, None

    def parse_direct_message(self, message_text):
        """
            Finds a direct mention (a mention that is at the beginning) in message text
            and returns the user ID which was mentioned. If there is no direct mention, returns None
        """
        matches = re.search(self.DIRECT_REGEX, message_text)
        # the first group contains the username, the second group contains the remaining message
        if matches:
            return matches.group(1), matches.group(2).strip()

        return None, None

    def ask_question(self, userid, question):
        self._questions += 1
        client_context = self.create_client_context(userid)
        response = client_context.bot.ask_question(client_context, question, responselogger=self)
        return self.renderer.render(client_context, response)

    def send_response(self, response, channel):
        # Sends the response back to the channel
        self._slack_client.api_call(
            "chat.postMessage",
            username=self.configuration.client_configuration.username,
            channel=channel,
            text=response or response
        )

    def handle_message(self, message, channel, userid):

        # Finds and executes the given message, filling in response
        response = self.ask_question(userid, message)
        YLogger.debug(self, "Slack sending [%s] to [%s]", message, userid)

        self.send_response(response, channel)

    def get_messages(self):
        return self._slack_client.rtm_read()

    def display_connected_message(self):
        outputLog(self, "Slack Bot connected and running...")

    def poll_and_answer(self):
        running = True
        try:
            messages = self.get_messages()
            if messages:
                self.parse_messages(messages)
            self.sleep(self._polling_interval)

        except KeyboardInterrupt:
            outputLog(self, "Slack client stopping via keyboard....")
            running = False

        except Exception as excep:
            YLogger.exception(self, "Failed to poll and answer", excep)

        return running


if __name__ == '__main__':

    outputLog(None, "Initiating Slack Client...")

    SLACK_CLIENT = SlackBotClient()
    SLACK_CLIENT.run()



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.utils.logging.ylogger import YLogger
from programy.storage.factory import StorageFactory


class BinariesManager:

    def __init__(self, binaries_configuration):
        assert binaries_configuration is not None

        self._configuration = binaries_configuration

    def _load_from_storage(self, storage_factory):
        storage_engine = storage_factory.entity_storage_engine(StorageFactory.BINARIES)
        binaries_storage = storage_engine.binaries_store()
        return binaries_storage.load_binary()

    def load_binary(self, storage_factory):
        try:
            if storage_factory.entity_storage_engine_available(StorageFactory.BINARIES) is True:
                YLogger.info(self, "Loading binary brain from [%s]", StorageFactory.BINARIES)

                return self._load_from_storage(storage_factory)

        except Exception as excep:
            YLogger.exception(self, "Failed to load binary file", excep)
            if self._configuration.load_aiml_on_binary_fail is False:
                raise excep

        return None

    def _save_to_storage(self, storage_factory, aiml_parser):
        storage_engine = storage_factory.entity_storage_engine(StorageFactory.BINARIES)
        binaries_storage = storage_engine.binaries_store()
        binaries_storage.save_binary(aiml_parser)

    def save_binary(self, storage_factory, aiml_parser):
        try:
            if storage_factory.entity_storage_engine_available(StorageFactory.BINARIES) is True:
                YLogger.info(self, "Saving binary brain to [%s]", StorageFactory.BINARIES)

                self._save_to_storage(storage_factory, aiml_parser)
                return True

        except Exception as failure:
            YLogger.info(self, "Failed to save binary [%s]", failure)

        return False


"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.config.container import BaseContainerConfigurationData
from programy.config.brain.overrides import BrainOverridesConfiguration
from programy.config.brain.defaults import BrainDefaultsConfiguration
from programy.config.brain.binaries import BrainBinariesConfiguration
from programy.config.brain.braintree import BrainBraintreeConfiguration
from programy.config.brain.services import BrainServicesConfiguration
from programy.config.brain.openchatbots import BrainOpenChatBotsConfiguration
from programy.config.brain.securities import BrainSecuritiesConfiguration
from programy.config.brain.dynamic import BrainDynamicsConfiguration
from programy.config.brain.tokenizer import BrainTokenizerConfiguration
from programy.config.brain.debugfiles import BrainDebugFilesConfiguration
from programy.utils.substitutions.substitues import Substitutions


class BrainConfiguration(BaseContainerConfigurationData):

    def __init__(self, section_name="brain"):
        self._overrides = BrainOverridesConfiguration()
        self._defaults = BrainDefaultsConfiguration()
        self._binaries = BrainBinariesConfiguration()
        self._braintree = BrainBraintreeConfiguration()
        self._services = BrainServicesConfiguration()
        self._openchatbots = BrainOpenChatBotsConfiguration()
        self._security = BrainSecuritiesConfiguration()
        self._dynamics = BrainDynamicsConfiguration()
        self._tokenizer = BrainTokenizerConfiguration()
        self._debugfiles = BrainDebugFilesConfiguration()
        BaseContainerConfigurationData.__init__(self, section_name)

    @property
    def overrides(self):
        return self._overrides

    @property
    def defaults(self):
        return self._defaults

    @property
    def binaries(self):
        return self._binaries

    @property
    def braintree(self):
        return self._braintree

    @property
    def services(self):
        return self._services

    @property
    def openchatbots(self):
        return self._openchatbots

    @property
    def security(self):
        return self._security

    @property
    def dynamics(self):
        return self._dynamics

    @property
    def tokenizer(self):
        return self._tokenizer

    @property
    def debugfiles(self):
        return self._debugfiles

    def check_for_license_keys(self, license_keys):
        self._overrides.check_for_license_keys(license_keys)
        self._defaults.check_for_license_keys(license_keys)
        self._binaries.check_for_license_keys(license_keys)
        self._braintree.check_for_license_keys(license_keys)
        self._services.check_for_license_keys(license_keys)
        self._openchatbots.check_for_license_keys(license_keys)
        self._security.check_for_license_keys(license_keys)
        self._dynamics.check_for_license_keys(license_keys)
        self._tokenizer.check_for_license_keys(license_keys)
        self._debugfiles.check_for_license_keys(license_keys)
        BaseContainerConfigurationData.check_for_license_keys(self, license_keys)

    def load_configuration(self, configuration_file, brain_config, bot_root, subs: Substitutions = None):
        if brain_config is not None:
            self._overrides.load_config_section(configuration_file, brain_config, bot_root, subs=subs)
            self._defaults.load_config_section(configuration_file, brain_config, bot_root, subs=subs)
            self._binaries.load_config_section(configuration_file, brain_config, bot_root, subs=subs)
            self._braintree.load_config_section(configuration_file, brain_config, bot_root, subs=subs)
            self._services.load_config_section(configuration_file, brain_config, bot_root, subs=subs)
            self._openchatbots.load_config_section(configuration_file, brain_config, bot_root, subs=subs)
            self._security.load_config_section(configuration_file, brain_config, bot_root, subs=subs)
            self._dynamics.load_config_section(configuration_file, brain_config, bot_root, subs=subs)
            self._tokenizer.load_config_section(configuration_file, brain_config, bot_root, subs=subs)
            self._debugfiles.load_config_section(configuration_file, brain_config, bot_root, subs=subs)

    def to_yaml(self, data, defaults=True):
            self.config_to_yaml(data, BrainOverridesConfiguration(), defaults)
            self.config_to_yaml(data, BrainDefaultsConfiguration(), defaults)
            self.config_to_yaml(data, BrainBinariesConfiguration(), defaults)
            self.config_to_yaml(data, BrainBraintreeConfiguration(), defaults)
            self.config_to_yaml(data, BrainServicesConfiguration(), defaults)
            self.config_to_yaml(data, BrainOpenChatBotsConfiguration(), defaults)
            self.config_to_yaml(data, BrainSecuritiesConfiguration(), defaults)
            self.config_to_yaml(data, BrainDynamicsConfiguration(), defaults)
            self.config_to_yaml(data, BrainTokenizerConfiguration(), defaults)
            self.config_to_yaml(data, BrainDebugFilesConfiguration(), defaults)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

# <sraix service="OPENCHAT"> @botname question </sraix>

from urllib.parse import quote
from programy.utils.logging.ylogger import YLogger
from programy.services.rest import GenericRESTService
from programy.config.brain.service import BrainServiceConfiguration
from programy.clients.restful.apikeys import APIKeysHandler
from programy.clients.restful.auth import RestBasicAuthorizationHandler
from programy.services.openchatbot.parser import OpenChatBotResponseParser
from programy.utils.console.console import outputLog


class OpenChatRESTService(GenericRESTService):

    def __init__(self, config: BrainServiceConfiguration, api=None):
        GenericRESTService.__init__(self, config, api)

    def _format_post_payload(self, client_context, question, lang=None, location=None):
        payload = {'query': question,
                   "userId": client_context.userid}

        if lang is not None:
            payload["lang"] = lang

        if location is not None:
            payload["location"] = location

        return payload

    def _format_get_url(self, url, client_context, question, lang=None, location=None, api_key=None):
        get_url = "%s?query=%s&userId=%s" % (url, quote(question), client_context.userid)

        if lang is not None:
            get_url = "%s&lang=%s" % (get_url, lang)

        if location is not None:
            get_url = "%s&location=%s" % (get_url, location)

        if api_key is not None:
            get_url = "%s&%s" % (get_url, APIKeysHandler.format_get_api_key_param(api_key))

        return get_url

    def _parse_response(self, client_context, text):
        parser = OpenChatBotResponseParser()
        parser.parse_response(text)

        if parser.status.code == 200:
            return parser.response.to_aiml()

        return client_context.bot.default_response

    def _parse_question(self, text):
        words = text.split()
        if len(words) < 2:
            raise Exception("Query test too short [%s]" % text)

        botname = words[0]

        query = " ".join(words[1:])

        return botname, query

    def _get_openchatbot(self, client_context, botname):

        openchatbot = client_context.brain.openchatbots.openchatbot(botname)

        if openchatbot is not None:
            return openchatbot

        raise Exception("No open chatbot available for [%s]" % botname)

    def ask_question(self, client_context, question: str):

        try:
            botname, query = self._parse_question(question)

            chatbot = self._get_openchatbot(client_context, botname)

            headers = {}

            if self.method == 'GET' and self.method in chatbot.methods:
                full_url = self._format_get_url(chatbot.url, client_context, query, api_key=chatbot.api_key)

                if chatbot.authorization == 'Basic':
                    RestBasicAuthorizationHandler.add_authorisation_header(client_context, headers)

                    response = self.api.get(full_url, headers=headers)

                else:
                    response = self.api.get(full_url)

            elif self.method == 'POST' and self.method in chatbot.methods:
                payload = self._format_post_payload(client_context, query)
                if chatbot.authorization == 'Basic' or chatbot.api_key is not None:

                    if chatbot.authorization == 'Basic':
                        RestBasicAuthorizationHandler.add_authorisation_header(client_context, headers)

                    if chatbot.api_key is not None:
                        APIKeysHandler.add_post_api_key_header(headers, chatbot.api_key)

                    response = self.api.post(chatbot.url, data=payload, headers=headers)
                else:
                    response = self.api.post(chatbot.url, data=payload)

            else:
                raise Exception("Unsupported REST method [%s]" % self.method)

            if response.status_code != 200:
                YLogger.error(client_context, "[%s] return status code [%d]", self.host, response.status_code)

            else:
                return self._parse_response(client_context, response.text)

        except Exception as excep:
            YLogger.exception(client_context, "Failed to resolve", excep)

        return ""


class OpenChatMetaBotRESTService(OpenChatRESTService):

    def __init__(self, config: BrainServiceConfiguration, api=None):
        OpenChatRESTService.__init__(self, config, api)

    def _parse_response(self, client_context, text):
        outputLog(self, text)
        return text



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import os
from programy.utils.logging.ylogger import YLogger
from programy.storage.stores.sql.store.sqlstore import SQLStore
from programy.storage.entities.lookups import LookupsStore
from programy.storage.stores.sql.dao.lookup import Denormal
from programy.storage.stores.sql.dao.lookup import Normal
from programy.storage.stores.sql.dao.lookup import Person
from programy.storage.stores.sql.dao.lookup import Person2
from programy.storage.stores.sql.dao.lookup import Gender
from programy.storage.entities.store import Store
from programy.mappings.base import DoubleStringPatternSplitCollection
from programy.utils.console.console import outputLog


class SQLLookupsStore(SQLStore, LookupsStore):

    def __init__(self, storage_engine):
        SQLStore.__init__(self, storage_engine)
        LookupsStore.__init__(self)

    def load_all(self, collector):
        self.load(collector)

    def load(self, collector, name=None):
        db_lookups = self._get_all()
        for db_lookup in db_lookups:
            key, value = DoubleStringPatternSplitCollection.process_key_value(db_lookup.key, db_lookup.value)
            collector.add_to_lookup(key, value)

    def _get_entity(self, key, value):
        raise NotImplementedError()  # pragma: no cover

    def _exists(self, key):
        raise NotImplementedError()  # pragma: no cover

    def add_to_lookup(self, key, value, overwrite_existing=False):
        lookup = self._exists(key)
        if lookup is not None:
            if overwrite_existing is True:
                YLogger.info(self, "Updating lookup in SQL [%s] [%s]", key, value)
                lookup.value = value

            else:
                YLogger.error(self, "Existing value in SQL lookup [%s] = [%s]", key, value)
                return False

        else:
            YLogger.debug(self, "Adding lookup to SQL [%s] = [%s]", key, value)
            lookup = self._get_entity(key=key, value=value)
            self._storage_engine.session.add(lookup)

        self._storage_engine.session.commit()
        return True

    def get_lookup(self):
        all = self._get_all()
        lookups = {}
        for keyvalue in all:
            lookups[keyvalue.key] = keyvalue.value
        return lookups

    def remove_lookup_key(self, key):
        raise NotImplementedError()         # pragma: no cover

    def remove_lookup(self):
        raise NotImplementedError()         # pragma: no cover

    def split_into_fields(self, line):
        return DoubleStringPatternSplitCollection.\
            split_line_by_pattern(line, DoubleStringPatternSplitCollection.RE_OF_SPLIT_PATTERN)

    def process_line(self, name, fields, verbose=False):
        if fields and len(fields) == 2:
            result = self.add_to_lookup(fields[0].upper(), fields[1].upper())
            if verbose is True:
                outputLog(self, "Key=[%s], Value={%s]" % (fields[0].upper(), fields[1].upper()))
            return result
        return False

    def _read_lookups_from_file(self, filename, verbose):
        count = 0
        success = 0
        with open(filename, "r") as lookup_file:
            for line in lookup_file:
                line = line.strip()
                fields = self.split_into_fields(line)
                if self.process_line(None, fields, verbose) is True:
                    success += 1
                count += 1
        return count, success

    def upload_from_file(self, filename, fileformat=Store.TEXT_FORMAT, commit=True, verbose=False):
        try:
            count, success = self._read_lookups_from_file(filename, verbose)
            self.commit(commit)
            return count, success

        except Exception as e:
            YLogger.exception(None, "Failed to upload lookups from file [%s]", e, filename)

        return 0, 0

class SQLDenormalStore(SQLLookupsStore):

    def __init__(self, storage_engine):
        SQLLookupsStore.__init__(self, storage_engine)

    def empty(self):
        self._get_all().delete()

    def remove_lookup_key(self, key):
        self._storage_engine.session.query(Denormal).filter(Denormal.key == key).delete()

    def remove_lookup(self):
        self._storage_engine.session.query(Denormal).delete()

    def _get_all(self):
        return self._storage_engine.session.query(Denormal)

    def _get_entity(self, key, value):
        return Denormal(key=key, value=value)

    def _exists(self, key):
        try:
            return self._storage_engine.session.query(Denormal).filter(Denormal.key == key).one()

        except Exception:
            return None


class SQLNormalStore(SQLLookupsStore):

    def __init__(self, storage_engine):
        SQLLookupsStore.__init__(self, storage_engine)

    def _get_all(self):
        return self._storage_engine.session.query(Normal)

    def empty(self):
        self._get_all().delete()

    def remove_lookup_key(self, key):
        self._storage_engine.session.query(Normal).filter(Normal.key == key).delete()

    def remove_lookup(self):
        self._storage_engine.session.query(Normal).delete()

    def _get_entity(self, key, value):
        return Normal(key=key, value=value)

    def _exists(self, key):
        try:
            return self._storage_engine.session.query(Normal).filter(Denormal.key == key).one()

        except Exception:
            return None


class SQLGenderStore(SQLLookupsStore):

    def __init__(self, storage_engine):
        SQLLookupsStore.__init__(self, storage_engine)

    def _get_all(self):
        return self._storage_engine.session.query(Gender)

    def empty(self):
        return self._get_all().delete()

    def remove_lookup_key(self, key):
        self._storage_engine.session.query(Gender).filter(Gender.key == key).delete()

    def remove_lookup(self):
        self._storage_engine.session.query(Gender).delete()

    def _get_entity(self, key, value):
        return Gender(key=key, value=value)

    def _exists(self, key):
        try:
            return self._storage_engine.session.query(Gender).filter(Gender.key == key).one()

        except Exception:
            return None


class SQLPersonStore(SQLLookupsStore):

    def __init__(self, storage_engine):
        SQLLookupsStore.__init__(self, storage_engine)

    def _get_all(self):
        return self._storage_engine.session.query(Person)

    def empty(self):
        return self._get_all().delete()

    def remove_lookup_key(self, key):
        self._storage_engine.session.query(Person).filter(Person.key == key).delete()

    def remove_lookup(self):
        self._storage_engine.session.query(Person).delete()

    def _get_entity(self, key, value):
        return Person(key=key, value=value)

    def _exists(self, key):
        try:
            return self._storage_engine.session.query(Person).filter(Person.key == key).one()

        except Exception:
            return None


class SQLPerson2Store(SQLLookupsStore):

    def __init__(self, storage_engine):
        SQLLookupsStore.__init__(self, storage_engine)

    def _get_all(self):
        return self._storage_engine.session.query(Person2)

    def empty(self):
        return self._get_all().delete()

    def remove_lookup_key(self, key):
        self._storage_engine.session.query(Person2).filter(Person2.key == key).delete()

    def remove_lookup(self):
        self._storage_engine.session.query(Person2).delete()

    def _get_entity(self, key, value):
        return Person2(key=key, value=value)

    def _exists(self, key):
        try:
            return self._storage_engine.session.query(Person2).filter(Denormal.key == key).one()

        except Exception:
            return None



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import datetime
import metoffer
from programy.utils.logging.ylogger import YLogger

DIRECTIONS = {'N': 'North',
              'NNE': 'North North East',
              'NE': 'North East',
              'ENE': 'East North East',
              'E': 'East',
              'ESE': 'East South East',
              'SE': 'South East',
              'SSE': 'South South East',
              'S': 'South',
              'SSW': 'South South West',
              'SW': 'South West',
              'WSW': 'West South West',
              'W': 'West',
              'WNW': 'West North West',
              'NW': 'North West',
              'NNW': 'North North West'}

PRESSURE_TENDANCY = {'F': "Falling",
                     'R': 'Rising'}


class DataPoint:

    def extract_attribute(self, json_data, name, data_type, time_period=None):
        if name in json_data:
            return json_data[name]
        else:
            if data_type == MetOfficeWeatherReport.OBSERVATION:
                YLogger.warning(self, '%s attribute missing from ObservationDataPoint data point', name)
            elif data_type == MetOfficeWeatherReport.FORECAST:
                if time_period == metoffer.THREE_HOURLY:
                    YLogger.warning(self, '%s attribute missing from three hourly forecast data point', name)
                if time_period == metoffer.DAILY:
                    YLogger.warning(self, '%s attribute missing from daily forecast data point', name)
            return None

    def direction_to_full_text(self, direction):
        if direction in DIRECTIONS:
            return DIRECTIONS[direction]
        return "Unknown"


class DailyForecastDayDataPoint(DataPoint):

    def __init__(self):
        self._type = None  # $
        self._wind_direction = None  # D
        self._temp_max = None  # Dm
        self._temperature_feels_like_max = None  # FDm
        self._wind_gust_noon = None  # Gn
        self._screen_relative_humidity_noon = None  # Hn
        self._precipitation_probability = None  # PPd
        self._wind_speed = None  # S
        self._uv_index_max = None  # U
        self._uv_guidance = None
        self._visibility_code = None  # V
        self._visibility_text = None
        self._weather_type_code = None  # W
        self._weather_type_text = None


    @property
    def type(self):
        return self._type

    #
    # Matches an AIML Pattern of
    # <pattern></pattern>
    #
    def to_program_y_text(self):
        text = "WEATHER TYPE %s WINDDIR %s WINDSPEED %s WINDGUST %s TEMP %s FEELS %s HUMID %s" \
               "RAINPROB %s" % (
                   self._type,
                   self._wind_direction,
                   self._wind_speed,
                   self._wind_gust_noon,
                   self._temp_max,
                   self._temperature_feels_like_max,
                   self._screen_relative_humidity_noon,
                   self._precipitation_probability)

        if self._uv_index_max is not None:
            text += " UVINDEX %s" % self._uv_index_max

        if self._uv_guidance is not None:
            text += " UVGUIDE %s" % self._uv_guidance

        if self._visibility_text is not None:
            text += " VIS %s" % self._visibility_text

        if self._weather_type_text is not None:
            text += " WEATHER %s" % self._weather_type_text

        return text

    def parse_json(self, json_data, data_type, time_period):
        self._type = self.extract_attribute(json_data, '$', data_type, time_period)
        self._wind_direction = self.extract_attribute(json_data, 'D', data_type, time_period)
        self._temp_max = self.extract_attribute(json_data, 'Dm', data_type, time_period)
        self._temperature_feels_like_max = self.extract_attribute(json_data, 'FDm', data_type, time_period)
        self._wind_gust_noon = self.extract_attribute(json_data, 'Gn', data_type, time_period)
        self._screen_relative_humidity_noon = self.extract_attribute(json_data, 'Hn', data_type, time_period)
        self._precipitation_probability = self.extract_attribute(json_data, 'PPd', data_type, time_period)
        self._wind_speed = self.extract_attribute(json_data, 'S', data_type, time_period)
        self._uv_index_max = self.extract_attribute(json_data, 'U', data_type, time_period)
        if self._uv_index_max is not None:
            self._uv_guidance = metoffer.guidance_UV(int(self._uv_index_max))
        self._visibility_code = self.extract_attribute(json_data, 'V', data_type, time_period)
        if self._visibility_code is not None:
            self._visibility_text = metoffer.VISIBILITY[self._visibility_code]
        self._weather_type_code = self.extract_attribute(json_data, 'W', data_type, time_period)
        if self._weather_type_code is not None:
            self._weather_type_text = metoffer.WEATHER_CODES[int(self._weather_type_code)]


class DailyForecastNightDataPoint(DataPoint):

    def __init__(self):
        self._type = None  # $
        self._wind_direction = None  # D
        self._temperature_feels_like_min = None  # FNm
        self._wind_gust_midnight = None  # Gm
        self._screen_relative_humidity_midnight = None  # Hm
        self._temp_min = None  # Nm
        self._precipitation_probability = None  # PPn
        self._wind_speed = None  # S
        self._visibility_code = None  # V
        self._visibility_text = None
        self._weather_type_code = None  # W
        self._weather_type_text = None

    @property
    def type(self):
        return self._type

    #
    # Matches an AIML Pattern of
    # <pattern></pattern>
    #
    def to_program_y_text(self):
        text = "WEATHER"

        if self._weather_type_text is not None:
            text += " TYPE %s" % self._weather_type_text

        text += " WINDDIR %s WINDGUST %s WINDSPEED %s TEMP %s FEELS %s HUMID %s RAINPROB %s" % (
                   self._wind_direction,
                   self._wind_gust_midnight,
                   self._wind_speed,
                   self._temp_min,
                   self._temperature_feels_like_min,
                   self._screen_relative_humidity_midnight,
                   self._precipitation_probability)

        if self._visibility_text is not None:
            text += " VISTEXT %s" % self._visibility_text

        if self._weather_type_text is not None:
            text += " WEATHER %s" % self._weather_type_text

        return text

    def parse_json(self, json_data, data_type, time_period):
        self._type = self.extract_attribute(json_data, '$', data_type, time_period)
        self._wind_direction = self.extract_attribute(json_data, 'D', data_type, time_period)
        self._temperature_feels_like_min = self.extract_attribute(json_data, 'FNm', data_type, time_period)
        self._wind_gust_midnight = self.extract_attribute(json_data, 'Gm', data_type, time_period)
        self._temp_min = self.extract_attribute(json_data, 'Nm', data_type, time_period)
        self._screen_relative_humidity_midnight = self.extract_attribute(json_data, 'Hm', data_type, time_period)
        self._precipitation_probability = self.extract_attribute(json_data, 'PPn', data_type, time_period)
        self._wind_speed = self.extract_attribute(json_data, 'S', data_type, time_period)
        self._visibility_code = self.extract_attribute(json_data, 'V', data_type, time_period)
        if self._visibility_code is not None:
            self._visibility_text = metoffer.VISIBILITY[self._visibility_code]
        self._weather_type_code = self.extract_attribute(json_data, 'W', data_type, time_period)
        if self._weather_type_code is not None:
            self._weather_type_text = metoffer.WEATHER_CODES[int(self._weather_type_code)]


class ThreeHourlyForecastDataPoint(DataPoint):

    def __init__(self):
        self._time = None  # $

        self._weather_type_code = None  # W
        self._weather_type_text = None

        self._temperature = None  # T
        self._temperature_feels_like = None  # F

        self._wind_gust = None  # G
        self._wind_direction = None  # D
        self._wind_direction_full = None
        self._wind_speed = None  # S

        self._visibility_code = None  # V
        self._visibility_text = None

        self._uv_index_max = None  # U
        self._uv_guidance = None

        self._precipitation_probability = None  # Pp
        self._screen_relative_humidity = None  # H

    @property
    def time(self):
        return self._time

    def parse_json(self, json_data, data_type, time_period):
        self._time = self.extract_attribute(json_data, '$', data_type, time_period)
        self._temperature_feels_like = self.extract_attribute(json_data, 'F', data_type, time_period)
        self._wind_gust = self.extract_attribute(json_data, 'G', data_type, time_period)
        self._screen_relative_humidity = self.extract_attribute(json_data, 'H', data_type, time_period)
        self._temperature = self.extract_attribute(json_data, 'T', data_type, time_period)
        self._visibility_code = self.extract_attribute(json_data, 'V', data_type, time_period)
        if self._visibility_code is not None:
            self._visibility_text = metoffer.VISIBILITY[self._visibility_code]
        self._wind_direction = self.extract_attribute(json_data, 'D', data_type, time_period)
        if self._wind_direction is not None:
            self._wind_direction_full = self.direction_to_full_text(self._wind_direction)
        self._wind_speed = self.extract_attribute(json_data, 'S', data_type, time_period)
        self._uv_index_max = self.extract_attribute(json_data, 'U', data_type, time_period)
        if self._uv_index_max is not None:
            self._uv_guidance = metoffer.guidance_UV(int(self._uv_index_max))
        self._weather_type_code = self.extract_attribute(json_data, 'W', data_type, time_period)
        if self._weather_type_code is not None:
            self._weather_type_text = metoffer.WEATHER_CODES[int(self._weather_type_code)]
            self._weather_type_text = self._weather_type_text.replace("(day)", "")
            self._weather_type_text = self._weather_type_text.replace("(night)", "")

        self._precipitation_probability = self.extract_attribute(json_data, 'Pp', data_type, time_period)

    #
    # Matches an AIML Pattern of
    # <pattern></pattern>
    #
    def to_program_y_text(self):
        text = "WEATHER"

        if self._weather_type_text is not None:
            text += " %s" % self._weather_type_text

        text += " TEMP %s FEELS %s WINDSPEED %s UVINDEX %s UVGUIDE %s RAINPROB %s HUMIDITY %s" % (
                   self._temperature,
                   self._temperature_feels_like,
                   self._wind_speed,
                   self._uv_index_max,
                   self._uv_guidance,
                   self._precipitation_probability,
                   self._screen_relative_humidity
               )

        if self._wind_direction is not None:
            text += " WINDDIR %s" % self._wind_direction

        if self._wind_direction_full is not None:
            text += " WINDDIRFULL %s" % self._wind_direction_full

        if self._visibility_text is not None:
            text += " VIS %s" % self._visibility_text

        return text


class ObservationDataPoint(DataPoint):

    def __init__(self):
        self._time = None  # $
        self._temperature = None  # T
        self._visibility = None  # V
        self._visibility_text = None
        self._wind_direction = None  # D
        self._wind_direction_full = None
        self._wind_speed = None  # S
        self._weather_type_code = None  # W
        self._weather_type_text = None
        self._pressure = None  # P
        self._pressure_tendancy = None  # Pt
        self._pressure_tendancy_full = None
        self._dew_point = None  # Dp
        self._screen_relative_humidity = None  # H

    @property
    def time(self):
        return self._time

    #
    # Matches an AIML Pattern of
    # <pattern>WEATHER TEMP * * HUMIDITY * * VISIBILITY V * VF * PRESSURE P * PT * PTF * WIND D * DF * S *</pattern>
    #
    def to_program_y_text(self):
        temp = self._temperature.split(".")
        humid = self._screen_relative_humidity.split(".")

        text = "WEATHER"
        if self._weather_type_text is not None:
            text += " %s" % self._weather_type_text

        text += " TEMP %s %s" % (temp[0], temp[1])

        text += " HUMIDITY %s %s" % (humid[0], humid[1])

        if self._visibility is not None:
            text += " VISIBILITY V %s VF %s" % (self._visibility, self._visibility_text)

        text += " PRESSURE P %s" % self._pressure
        if self._pressure_tendancy is not None:
            text += " PT %s PTF %s" % (self._pressure_tendancy, self._pressure_tendancy_full)

        if self._wind_direction is not None:
            text += " WIND D %s DF %s" % (self._wind_direction, self._wind_direction_full)

        text += " S %s" % self._wind_speed

        return text

    def parse_json(self, json_data, data_type, time_period):
        self._time = self.extract_attribute(json_data, '$', data_type, time_period)
        self._temperature = self.extract_attribute(json_data, 'T', data_type, time_period)
        self._visibility = self.extract_attribute(json_data, 'V', data_type, time_period)
        if self._visibility is not None:
            self._visibility_text = self.parse_visibility_to_text(self._visibility)
        self._wind_direction = self.extract_attribute(json_data, 'D', data_type, time_period)
        if self._wind_direction is not None:
            self._wind_direction_full = self.direction_to_full_text(self._wind_direction)
        self._wind_speed = self.extract_attribute(json_data, 'S', data_type, time_period)
        self._weather_type_code = self.extract_attribute(json_data, 'W', data_type, time_period)
        if self._weather_type_code is not None:
            self._weather_type_text = metoffer.WEATHER_CODES[int(self._weather_type_code)]
        self._pressure = self.extract_attribute(json_data, 'P', data_type, time_period)
        self._pressure_tendancy = self.extract_attribute(json_data, 'Pt', data_type, time_period)
        if self._pressure_tendancy is not None:
            self._pressure_tendancy_full = self.parse_pressure_tendancy(self._pressure_tendancy)
        self._dew_point = self.extract_attribute(json_data, 'Dp', data_type, time_period)
        self._screen_relative_humidity = self.extract_attribute(json_data, 'H', data_type, time_period)

    def parse_visibility_to_text(self, code):
        distance_in_kms = int(code) // 1000
        if distance_in_kms < 1:
            return "Very poor"
        elif distance_in_kms > 0 and distance_in_kms < 4:
            return "Poor"
        elif distance_in_kms > 3 and distance_in_kms < 10:
            return "Moderate"
        elif distance_in_kms > 9 and distance_in_kms < 20:
            return "Good"
        elif distance_in_kms > 19 and distance_in_kms < 40:
            return "Very Good"
        else: # > 39:
            return "Excellent"

    def parse_pressure_tendancy(self, tendancy):
        if tendancy in PRESSURE_TENDANCY:
            return PRESSURE_TENDANCY[tendancy]
        return "Unknown"


class Report:

    def __init__(self, data_type, time_period):
        self._data_type = data_type
        self._time_period = time_period

        self._time_periods = []

        self._type = None
        self._report_date = None

    @property
    def time_period(self):
        return self._time_period

    @property
    def time_periods(self):
        return self._time_periods

    @property
    def type(self):
        return self._type

    @property
    def report_date(self):
        return self._report_date

    def parse_json(self, json_data):
        for element in json_data['Rep']:
            if self._data_type == MetOfficeWeatherReport.OBSERVATION:
                period = ObservationDataPoint()

            elif self._data_type == MetOfficeWeatherReport.FORECAST:
                if self._time_period == metoffer.DAILY:
                    if element['$'] == 'Day':
                        period = DailyForecastDayDataPoint()

                    elif element['$'] == 'Night':
                        period = DailyForecastNightDataPoint()

                    else:
                        raise ValueError("Unknown report type %s" % element['$'])

                elif self._time_period == metoffer.THREE_HOURLY:
                    period = ThreeHourlyForecastDataPoint()

                else:
                    raise ValueError("Unknown time period %s" % self._time_period)

            else:
                raise ValueError("Unknown data type %s" % self._data_type)

            period.parse_json(element, self._data_type, self._time_period)
            self._time_periods.append(period)

        if self._data_type == MetOfficeWeatherReport.OBSERVATION or self._time_period == metoffer.THREE_HOURLY:
            self._time_periods.sort(key=lambda period: int(period.time))

        if 'type' in json_data:
            self._type = json_data['type']
        else:
            raise ValueError("type missing from DV data")

        if 'value' in json_data:
            try:
                self._report_date = datetime.datetime.strptime(json_data['value'], "%Y-%m-%dZ")
            except:
                raise ValueError("Invalid date time format for value in Report data")
        else:
            raise ValueError("value missing from Report data")

    def get_last_time_period(self):
        if self._time_periods:
            return self._time_periods[-1]
        return None

    def get_time_period_by_type(self, period_type):
        if self._time_periods:
            for time_period in self._time_periods:
                if time_period.type == period_type:
                    return time_period
        return None

    def get_time_period_by_time(self, time):
        if self._time_periods:
            for time_period in self._time_periods:
                if time_period.time == time:
                    return time_period
        return None


class Location:

    def __init__(self, data_type, time_period):
        self._data_type = data_type
        self._time_period = time_period
        self._reports = []

        self._continent = None
        self._country = None
        self._elevation = None
        self._i = None
        self._lat = None
        self._lon = None
        self._name = None

    @property
    def time_period(self):
        return self._time_period

    @property
    def data_type(self):
        return self._data_type

    @property
    def reports(self):
        return self._reports

    def get_latest_report(self):
        if self._reports:
            return self._reports[-1]
        return None

    def get_report_for_date(self, report_date):
        if self._reports:
            for report in self._reports:
                if report.report_date == report_date:
                    return report
        return None

    def parse_json(self, json_data):
        for element in json_data['Period']:
            report = Report(self._data_type, self._time_period)
            report.parse_json(element)
            self._reports.append(report)

        # Sort Reports by _report_date which is is either
        # a) The date of the observation
        # b) The date of the daily forecast
        if self._reports:
            self._reports.sort(key=lambda report: report.report_date)

        if 'continent' in json_data:
            self._continent = json_data['continent']
        else:
            raise ValueError("continent missing from Location data")

        if 'country' in json_data:
            self._country = json_data['country']
        else:
            raise ValueError("country missing from Location data")

        if 'elevation' in json_data:
            self._elevation = json_data['elevation']
        else:
            raise ValueError("elevation missing from Location data")

        if 'i' in json_data:
            self._i = json_data['i']
        else:
            raise ValueError("i missing from Location data")

        if 'lat' in json_data:
            self._lat = json_data['lat']
        else:
            raise ValueError("lat missing from Location data")

        if 'lon' in json_data:
            self._lon = json_data['lon']
        else:
            raise ValueError("lon missing from Location data")

        if 'name' in json_data:
            self._name = json_data['name']
        else:
            raise ValueError("name missing from Location data")


class DV:

    def __init__(self, data_type, time_period):
        self._data_type = data_type
        self._time_period = time_period
        self._date = None
        self._type = None
        self._location = None

    @property
    def location(self):
        return self._location

    def parse_json(self, json_data):

        if "Location" not in json_data:
            raise ValueError("Location missing from DV data")
        self._location = Location(self._data_type, self._time_period)
        self._location.parse_json(json_data['Location'])

        if 'dataDate' in json_data:
            try:
                self._date = datetime.datetime.strptime(json_data['dataDate'], "%Y-%m-%dT%H:%M:%SZ")
            except:
                raise ValueError("Invalid date format for dataDate in DV data")
        else:
            raise ValueError("dataDate missing from DV data")

        if 'type' in json_data:
            self._type = json_data['type']
        else:
            raise ValueError("type missing from DV data")


class SiteReport:

    def __init__(self, data_type, time_period):
        self._time_period = time_period
        self._data_type = data_type
        self._dv = None

    @property
    def dv(self):
        return self._dv

    def parse_json(self, json_data):
        if "DV" not in json_data:
            raise ValueError("DV missing from site report data")
        self._dv = DV(self._data_type, self._time_period)
        self._dv.parse_json(json_data['DV'])


class MetOfficeWeatherReport:
    FORECAST = 1
    OBSERVATION = 2

    def __init__(self, data_type, time_period=0):
        self._data_type = data_type
        self._time_period = time_period
        self._site_report = None
        self._time_periods = {}

    def parse_json(self, json_data):
        if "SiteRep" not in json_data:
            raise ValueError("SiteRep missing from weather report")
        else:
            self._site_report = SiteReport(self._data_type, self._time_period)
            self._site_report.parse_json(json_data["SiteRep"])

        self.cache_time_periods()

    def cache_time_periods(self):
        pass                        # pragam: no cover


class MetOfficeObservation(MetOfficeWeatherReport):

    def __init__(self):
        MetOfficeWeatherReport.__init__(self, MetOfficeWeatherReport.OBSERVATION)

    def cache_time_periods(self):
        self._time_periods.clear()
        for report in self._site_report.dv.location.reports:
            for time_period in report.time_periods:
                fulltime = report.report_date + datetime.timedelta(minutes=int(time_period.time))
                self._time_periods[fulltime] = time_period

    def get_observations(self):
        return self._time_periods

    def get_latest_observation(self):
        items = list(self._time_periods.items())
        return items[-1][1]

    def get_last_n_observations(self, n):
        return list(self._time_periods.items())[-n:]


class MetOfficeForecast(MetOfficeWeatherReport):

    def __init__(self, time_period):
        MetOfficeWeatherReport.__init__(self, MetOfficeWeatherReport.FORECAST, time_period)

    def get_forecasts(self):
        return self._time_periods

    def get_latest_forecast(self):
        items = list(self._time_periods.items())
        return items[-1][1]


class MetOffice5DayForecast(MetOfficeForecast):

    def __init__(self):
        MetOfficeForecast.__init__(self, metoffer.DAILY)

    def cache_time_periods(self):
        self._time_periods.clear()
        for report in self._site_report.dv.location.reports:
            daytime = report.report_date + datetime.timedelta(hours=6)
            self._time_periods[daytime] = report.time_periods[0]
            nighttime = report.report_date + datetime.timedelta(hours=18)
            self._time_periods[nighttime] = report.time_periods[1]

    def _calc_date_n_days_ahead(self, days, fromdate=None):
        if fromdate is None:
            return datetime.datetime.now() + datetime.timedelta(days=days)
        else:
            return datetime.datetime.strptime(fromdate, '%Y-%m-%dZ') + datetime.timedelta(days=days)

    def get_forecast_for_n_days_ahead(self, days, fromdate=None):
        search_date = self._calc_date_n_days_ahead(days, fromdate=fromdate)

        for day_forecast in self._time_periods.keys():
            if day_forecast > search_date:
                return self._time_periods[day_forecast].to_program_y_text()

        return None


class MetOffice24HourForecast(MetOfficeForecast):

    def __init__(self):
        MetOfficeForecast.__init__(self, metoffer.THREE_HOURLY)

    def cache_time_periods(self):
        self._time_periods.clear()
        for report in self._site_report.dv.location.reports:
            for time_period in report.time_periods:
                period_datetime = report.report_date + datetime.timedelta(minutes=int(time_period.time))
                self._time_periods[period_datetime] = time_period

    def _calc_date_n_hours_ahead(self, hours, fromdate=None):
        if fromdate is None:
            return datetime.datetime.now() + datetime.timedelta(hours=hours)
        else:
            return datetime.datetime.strptime(fromdate,  "%Y-%m-%dZ") + datetime.timedelta(hours=hours)

    def get_forecast_for_n_hours_ahead(self, hours, fromdate=None):
        search_date = self._calc_date_n_hours_ahead(hours, fromdate=fromdate)

        for hour_forecast in self._time_periods.keys():
            if hour_forecast > search_date:
                return self._time_periods[hour_forecast].to_program_y_text()

        return None


class MetOffice:

    def __init__(self):
        self._met_office_api = None

    def check_for_license_keys(self, license_keys):

        if license_keys.has_key('METOFFICE_API_KEY'):
            api_key = license_keys.get_key('METOFFICE_API_KEY')
        else:
            raise Exception("No valid license key METOFFICE_API_KEY found")

        self._met_office_api = metoffer.MetOffer(api_key)

    def get_forecast_data(self, lat, lon, forecast_type):
        return self._met_office_api.nearest_loc_forecast(lat, lon, forecast_type)       # pragma: no cover

    def get_observation_data(self, lat, lon):
        return self._met_office_api.nearest_loc_obs(lat, lon)                           # pragma: no cover

    def nearest_location_forecast(self, lat, lon, forecast_type):
        if forecast_type == metoffer.THREE_HOURLY:
            forecast = MetOffice24HourForecast()

        elif forecast_type == metoffer.DAILY:
            forecast = MetOffice5DayForecast()

        else:
            raise ValueError("Unsupported forecast_type: [%s]" % forecast_type)

        json_data = self.get_forecast_data(lat, lon, forecast_type)
        forecast.parse_json(json_data)
        return forecast

    def twentyfour_hour_forecast(self, lat, lon):
        return self.nearest_location_forecast(lat, lon, metoffer.THREE_HOURLY)

    def five_day_forecast(self, lat, lon):
        return self.nearest_location_forecast(lat, lon, metoffer.DAILY)

    def nearest_location_observation(self, lat, lon):
        report = MetOfficeObservation()
        json_data = self.get_observation_data(lat, lon)
        report.parse_json(json_data)
        return report

    def current_observation(self, lat, lon):
        return self.nearest_location_observation(lat, lon)



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from programy.parser.template.nodes.base import TemplateNode
from programy.utils.text.text import TextUtils
from programy.parser.exceptions import ParserException


class TemplateListNode(TemplateNode):

    def __init__(self):
        TemplateNode.__init__(self)
        self._items = []

    def resolve_list_items(self, client_context):
        resolved = ""
        for item in self._items:
            resolved += "<item>%s</item>" % item.resolve(client_context)
        return resolved

    def resolve_to_string(self, client_context):
        resolved = "<list>"
        resolved += self.resolve_list_items(client_context)
        resolved += "</list>"
        return resolved

    def to_string(self):
        return "[LIST %d]" % (len(self._items))

    def to_xml(self, client_context):
        return self.resolve_to_string(client_context)

    #######################################################################################################
    #

    def parse_expression(self, graph, expression):
        head_text = self.get_text_from_element(expression)
        self.parse_text(graph, head_text)

        for child in expression:
            tag_name = TextUtils.tag_from_text(child.tag)

            if tag_name == 'item':
                item = self.parse_children_as_word_node(graph, child)
                self._items.append(item)
            else:
                raise ParserException("Invalid child node in list")

            tail_text = self.get_tail_from_element(child)
            self.parse_text(graph, tail_text)

        if len(self._items) == 0:
            raise ParserException("No items in list")


import sys
import xml.etree.ElementTree as ET
import csv
import re

class AIMLToCSVGenerator(object):

    def __init__(self):
        self._input = sys.argv[1]
        self._output = sys.argv[2]

    def run(self):
        csv_file = None

        try:
            tree = ET.parse(self._input)
            csv_file = open(self._output, "w+")
            csv_writer = csv.writer(csv_file, delimiter=',', quotechar='"', quoting=csv.QUOTE_ALL)
            root = tree.getroot()
            for element in root:
                if element.tag == 'category':
                    self.parse_category_to_file(csv_writer, element, topic="*")
                elif element.tag == 'topic':
                    for child in element:
                        if child.tag == 'category':
                            self.parse_category_to_file(csv_writer, child, topic=element.attrib['name'])

        except Exception as excep:
            print (excep)
        finally:
            if csv_file is not None:
                csv_file.flush()
                csv_file.close ()

    def parse_category_to_file(self, csv_writer, category, topic):
        pattern = None
        that = "*"
        template = None
        for element in category:
            if element.tag == 'pattern':
                pattern = AIMLToCSVGenerator.element_to_string(element)
            elif element.tag == 'that':
                that = element.text
            elif element.tag == 'template':
                template = AIMLToCSVGenerator.element_to_string(element)

        pattern = AIMLToCSVGenerator.strip_all_whitespace(pattern)
        template = AIMLToCSVGenerator.strip_all_whitespace(template)
        topic = AIMLToCSVGenerator.strip_all_whitespace(topic)
        that = AIMLToCSVGenerator.strip_all_whitespace(that)

        csv_writer.writerow([pattern, topic, that, template])

    @staticmethod
    def element_to_string(element):
        s = element.text or ""
        for sub_element in element:
            s += ET.tostring(sub_element).decode("utf-8")
        s += element.tail
        return s

    @staticmethod
    def strip_all_whitespace(string):
        first_pass = re.sub(r'[\n\t\r+]', '', string)
        second_pass = re.sub(r'\s+', ' ', first_pass)
        return second_pass.strip()

if __name__ == '__main__':

    def run():
        print("Convertin AIML to CSV...")
        generator = AIMLToCSVGenerator()
        generator.run()

    run()



"""
Copyright (c) 2016-2020 Keith Sterling http://www.keithsterling.com

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import json
import os
from flask import Flask, request, jsonify
from programy.utils.logging.ylogger import YLogger
from programy.clients.restful.flask.client import FlaskRestBotClient
from programy.clients.restful.flask.alexa.config import AlexaConfiguration
from programy.utils.console.console import outputLog


class AlexaBotClient(FlaskRestBotClient):

    def __init__(self, argument_parser=None):
        FlaskRestBotClient.__init__(self, 'alexa', argument_parser)

        YLogger.debug(self, "Alexa Client is running....")

        self._load_intent_mappings()

        outputLog(self, "Alexa Client loaded")

    def _load_intent_mappings(self):
        try:
            if self.configuration.client_configuration.intent_map_file is None:
                filename = os.path.dirname(__file__) + os.sep + "intents.maps"
            else:
                filename = self.configuration.client_configuration.intent_map_file

            with open(filename, 'r') as myfile:
                data = myfile.read()

            self._intent_mappings = json.loads(data)

        except Exception as excep:
            YLogger.exception_nostack(self, "Failed to load intent mappings", excep)
            self._intent_mappings = {}

    def get_client_configuration(self):
        return AlexaConfiguration()

    def _ask_question(self, client_context, question):
        reply = ""
        try:
            self._questions += 1
            response = client_context.bot.ask_question(client_context, question, responselogger=self)
            reply = self.renderer.render(client_context, response)

        except Exception as e:
            YLogger.exception(client_context, "Error getting reply from bot", e)

        return reply

    def _to_json(self, data):
        return jsonify(data)

    def _create_response(self, response, responsetype="PlainText", playBehavior="REPLACE_ENQUEUED",
                         shouldEndSession=False):
        reply = {}
        reply["version"] = "1.0"
        reply["response"] = {}
        if reply is not None:
            ssml_reply = "<speak>%s</speak>" % response
            reply["response"]["outputSpeech"] = {}
            reply["response"]["outputSpeech"]["type"] = responsetype
            reply["response"]["outputSpeech"]["text"] = response
            reply["response"]["outputSpeech"]["ssml"] = ssml_reply
            reply["response"]["outputSpeech"]["playBehavior"] = playBehavior
        reply["response"]["shouldEndSession"] = shouldEndSession
        return self._to_json(reply)

    def _extract_question(self, intent):
        if 'slots' in intent:
            slots = intent['slots']

            if 'text' in slots:
                text_slot = slots['text']

                if 'value' in text_slot:
                    return text_slot['value']

        return ""

    def _add_intent(self, intent_name, value):
        if intent_name in self._intent_mappings:
            return "%s %s" % (self._intent_mappings[intent_name], value)

        return value

    def _should_leave(self, intent_name):
        if intent_name in self.configuration.client_configuration.leave_intents:
            return True
        return False

    def get_reply_from_bot(self, client_context, text, srai):
        if srai is not None:
            reply = self._ask_question(client_context, srai)
            if reply is None or reply == "":
                reply = text
        else:
            reply = text

        return reply

    def _handle_launch_request(self, client_context):
        if self.configuration.client_configuration.launch_srai is not None:
            reply = self._ask_question(client_context, self.configuration.client_configuration.launch_srai)

        else:
            reply = self.configuration.client_configuration.launch_text

        return self._create_response(reply)

    def _handle_intent_request(self, client_context, request):

        if 'intent' not in request:
            raise Exception("Invalid request, intent missing!")

        intent = request['intent']

        if 'name' not in intent:
            raise Exception("Invalid intent, name missing!")

        intent_name = intent['name']

        if intent_name == 'AMAZON.CancelIntent':
            return self._handle_cancel_request(client_context, self._should_leave(intent_name))

        elif intent_name == 'AMAZON.StopIntent':
            return self._handle_stop_request(client_context, self._should_leave(intent_name))

        elif intent_name == 'AMAZON.HelpIntent':
            return self._handle_help_request(client_context)

        else:
            value = self._extract_question(intent)

            question = self._add_intent(intent_name, value)

            return self._handle_reply_request(client_context, question)

    def _handle_leave_request(self):
        return self._create_response("See you later, aligator", shouldEndSession=True)

    def _handle_reply_request(self, client_context, question):
        reply = self._ask_question(client_context, question)
        return self._create_response(reply)

    def _handle_cancel_request(self, client_context, shouldEndSession=False):
        reply = self.get_reply_from_bot(client_context, self.configuration.client_configuration.cancel_text,
                                        self.configuration.client_configuration.cancel_srai)
        return self._create_response(reply, shouldEndSession=shouldEndSession)

    def _handle_stop_request(self, client_context, shouldEndSession=False):
        reply = self.get_reply_from_bot(client_context, self.configuration.client_configuration.stop_text,
                                        self.configuration.client_configuration.stop_srai)
        return self._create_response(reply, shouldEndSession=shouldEndSession)

    def _handle_help_request(self, client_context):
        reply = self.get_reply_from_bot(client_context, self.configuration.client_configuration.help_text,
                                        self.configuration.client_configuration.help_srai)
        return self._create_response(reply)

    def _handle_error(self, client_context):
        reply = self.get_reply_from_bot(client_context, self.configuration.client_configuration.error_text,
                                        self.configuration.client_configuration.error_srai)
        return self._create_response(reply)

    def _get_userid(self, request):
        if 'session' not in request:
            raise Exception("Invalid request, session missing")

        if "user" not in request['session']:
            raise Exception("Invalid session, user missing")

        if "userId" not in request['session']['user']:
            raise Exception("Invalid user, userId missing")

        return request['session']['user']['userId']

    def receive_message(self, http_request):

        skill_json = http_request.json

        recipient_id = self._get_userid(skill_json)

        if 'request' not in skill_json:
            raise Exception("Invalid HTTP request, request missing!")

        skill_request = skill_json['request']

        if 'type' not in skill_request:
            raise Exception("Invalid skill request, type missing!")

        request_type = skill_request['type']

        client_context = self.create_client_context(recipient_id)

        try:
            if request_type == 'LaunchRequest':
                return self._handle_launch_request(client_context)

            elif request_type == 'IntentRequest':
                return self._handle_intent_request(client_context, skill_request)

            elif request_type == 'SessionEndedRequest':
                return self._handle_stop_request(client_context, shouldEndSession=True)

            else:
                raise Exception("Unknown/Unhandled request")

        except Exception as excep:
            YLogger.exception(client_context, "Unknown/Unhandled request [%s]", request_type, excep)
            return self._handle_error(client_context)


if __name__ == "__main__":

    outputLog(None, "Initiating Alexa Client...")

    ALEXA_CLIENT = AlexaBotClient()

    APP = Flask(__name__)


    @APP.route(ALEXA_CLIENT.configuration.client_configuration.api, methods=['GET', 'POST'])
    def receive_message():
        try:
            return ALEXA_CLIENT.receive_message(request)

        except Exception as e:
            outputLog(None, "Alexa failed receiving message", e)
            YLogger.exception(None, "Alexa Error", e)


    ALEXA_CLIENT.run(APP)
